from Utility import Utility
input = Utility.input
class Raga:
    """Ragas are like Changes, but they have two directions,
    with possible multiple options per direction
    They also have articulations that may occur before, during, or after notes.
    They may have Tala (rhythmic figure[s]) associated with it
    Melkarta Ragas (Melas) have a mela number associated with them
    Finally, they will have a name, and possible alternate names"""

    # https: // asmp - eurasipjournals.springeropen.com / articles / 10.1186 / s13636 - 016 - 00    83 - z  # Fig2
    melaNameByMelaNumber = []
    melaNameByMelaNumber.insert(0, "Mela Error")
    melaNameByMelaNumber.insert(1, "Kanakangi")
    melaNameByMelaNumber.insert(2, "Ratnangi")
    melaNameByMelaNumber.insert(3, "Ganamurti")
    melaNameByMelaNumber.insert(4, "Vanaspati,Venaspati")
    melaNameByMelaNumber.insert(5, "Manavati,Tenarupi")
    melaNameByMelaNumber.insert(6, "Tanarupi,Tenarupi")
    melaNameByMelaNumber.insert(7, "Senavati")
    melaNameByMelaNumber.insert(8, "Hanumatodi,Hanumattodi")
    melaNameByMelaNumber.insert(9, "Dhenuka")
    melaNameByMelaNumber.insert(10, "Natakapriya")
    melaNameByMelaNumber.insert(11, "Kokilapriya")
    melaNameByMelaNumber.insert(12, "Rupavati")
    melaNameByMelaNumber.insert(13, "Gayakapriya")
    melaNameByMelaNumber.insert(14, "Vakulabharanam")
    melaNameByMelaNumber.insert(15, "Mayamalavagowla,Mayamalavagaula")
    melaNameByMelaNumber.insert(16, "Chakravakam,Cakravka")
    melaNameByMelaNumber.insert(17, "Suryakantam")
    melaNameByMelaNumber.insert(18, "Hatakambari")
    melaNameByMelaNumber.insert(19, "Jhankaradhwani")
    melaNameByMelaNumber.insert(20, "Natabhairavi")
    melaNameByMelaNumber.insert(21, "Keeravani")
    melaNameByMelaNumber.insert(22, "Kharaharapriya")
    melaNameByMelaNumber.insert(23, "Gourimanohari")
    melaNameByMelaNumber.insert(24, "Varunapriya")
    melaNameByMelaNumber.insert(25, "Mararanjani")
    melaNameByMelaNumber.insert(26, "Charukesi")
    melaNameByMelaNumber.insert(27, "Sarasangi")
    melaNameByMelaNumber.insert(28, "Harikambhoji")
    melaNameByMelaNumber.insert(
        29, "Dheerasankarabaranam,Shankarabharanam,Dhirasankarabharana"
    )
    melaNameByMelaNumber.insert(30, "Naganandini,Ragavardhini")
    melaNameByMelaNumber.insert(31, "Yagapriya")
    melaNameByMelaNumber.insert(32, "Ragavardhini")
    melaNameByMelaNumber.insert(33, "Gangeyabhushani")
    melaNameByMelaNumber.insert(
        34, "Vagadheeswari,Vagadhisvari,Vagadheeswari,Vagadhibhusani"
    )
    melaNameByMelaNumber.insert(35, "Shulini,Sulini")
    melaNameByMelaNumber.insert(36, "Chalanata")
    melaNameByMelaNumber.insert(37, "Salagam,Salaga")
    melaNameByMelaNumber.insert(38, "Jalarnavam,Jalarnava")
    melaNameByMelaNumber.insert(39, "Jhalavarali,Jhalavarli")
    melaNameByMelaNumber.insert(40, "Navaneetam,Navanitam")
    melaNameByMelaNumber.insert(41, "Pavani")
    melaNameByMelaNumber.insert(42, "Raghupriya")
    melaNameByMelaNumber.insert(43, "Gavambhodi,Gavambodhi")
    melaNameByMelaNumber.insert(44, "Bhavapriya")
    melaNameByMelaNumber.insert(
        45, "Shubhapantuvarali,Subhapantuvarali,Shubhapanturavali"
    )
    melaNameByMelaNumber.insert(46, "Shadvidamargini,Shubhapanturavali")
    melaNameByMelaNumber.insert(47, "Suvarnangi")
    melaNameByMelaNumber.insert(48, "Divyamani,Dvyamani")
    melaNameByMelaNumber.insert(49, "Dhavalambari")
    melaNameByMelaNumber.insert(50, "Namanarayani")
    melaNameByMelaNumber.insert(51, "Kamavardani,Kamavardhani")
    melaNameByMelaNumber.insert(52, "Ramapriya")
    melaNameByMelaNumber.insert(53, "Gamanashrama,Gamanasrama")
    melaNameByMelaNumber.insert(54, "Vishwambari,Visvambhari")
    melaNameByMelaNumber.insert(55, "Shamalangi,Syamalangi")
    melaNameByMelaNumber.insert(56, "Shanmukhapriya,Sanmukhapriya")
    melaNameByMelaNumber.insert(57, "Simhendramadhyamam,Simhendramadhyama")
    melaNameByMelaNumber.insert(58, "Hemavati")
    melaNameByMelaNumber.insert(59, "Dharmavati")
    melaNameByMelaNumber.insert(60, "Neetimati,Nitimati")
    melaNameByMelaNumber.insert(61, "Kantamani")
    melaNameByMelaNumber.insert(62, "Rishabhapriya,Risabhapriya")
    melaNameByMelaNumber.insert(63, "Latangi")
    melaNameByMelaNumber.insert(64, "Vachaspati,Vacaspati")
    melaNameByMelaNumber.insert(65, "Mechakalyani,Mechakalyani,Mecakalyani")
    melaNameByMelaNumber.insert(66, "Chitrambari,Chitrambari,Citrambari")
    melaNameByMelaNumber.insert(67, "Sucharitra,Sucaritra")
    melaNameByMelaNumber.insert(68, "Jyoti swarupini,Jyotisvarupini")
    melaNameByMelaNumber.insert(69, "Dhatuvardani,Dhatuvardhani")
    melaNameByMelaNumber.insert(70, "Nasikabhushani,Nasikabhusani")
    melaNameByMelaNumber.insert(71, "Kosalam")
    melaNameByMelaNumber.insert(72, "Rasikapriya")

    @classmethod
    def makeFestivalSounds(
        cls,
        includeAssociatedChords=True,
        includeAssociatedModes=False,
        sortByPageNumber=True,
        showDebug=False,
    ) -> [int]:
        """returns subChange numbers of melaNameByMelaNumber and associated changes."""
        _melaPages = [Raga.pageFromMelaNumber[i] for i in range(73)][1:]
        _chordPages = []


        _numOfDistinctChordsIncluded = 2
        if includeAssociatedChords:
            for melaPage in _melaPages:
                from .Change import Change
                _mela = Change.makeFromChangeNumber(melaPage)
                if showDebug:
                    print(
                        "Mela: {}\nnum{}\nmelaP{}".format(
                            _mela,
                            _numOfDistinctChordsIncluded,
                            _mela.getDistinctChordTypes(),
                        )
                    )
                _chordTypes = min(
                    _numOfDistinctChordsIncluded, len(_mela.getDistinctChordTypes())
                )

                for chordType in range(_chordTypes):
                    for note in range(len(_mela)):
                        _chord = _mela.getDistinctScaleChord(
                            note, chordType, returnChordQuality=False
                        )
                        if showDebug:
                            print("as " + str(_chord))
                        _chordPages.append(
                            _chord.getChangeNumber(decorateChapter=False)
                        )
        _chordPages = list(set(_chordPages))
        if showDebug:
            input(
                "Mela Pages {} \nCHord Pages{}".format(
                    [Change.makeFromChangeNumber(i) for i in _melaPages], _chordPages
                )
            )

        if includeAssociatedModes:
            _modePages = []
            for p in _melaPages:
                _mela = Change.makeFromChangeNumber(p)
                for n in range(len(_mela.notes)):
                    _mode = _mela.mode(n)
                    _modePage = _mode.getChangeNumber(decorateChapter=False)
                    if not _modePage in _modePages:
                        _modePages.append(_modePage)
            _melaPages = _melaPages + _modePages
        if includeAssociatedChords:
            _melaPages = _melaPages + _chordPages
            if showDebug:
                print("Chord Pages", _chordPages)

        _melaPages = list(set(_melaPages))
        if sortByPageNumber:
            _melaPages.sort()
        return _melaPages

    @classmethod
    def shortenRagaNameWithSymbols(
        cls,
        ragaName: str,
        replaceDirections=True,
        replaceRagaAndMela=True,
        useSpaces=True,
    ) -> str:
        if useSpaces:
            _space = " "
        else:
            _space = ""
        if replaceDirections:
            ragaName = ragaName.replace(
                " (ascending)", _space + Unicode.chars["Ascending"]
            ).replace(" (descending)", _space + Unicode.chars["Descending"])
        if replaceRagaAndMela:
            ragaName = ragaName.replace(
                "Raga ", Unicode.chars["Raga"] + _space
            ).replace("Mela ", Unicode.chars["Mela"] + _space)
        return ragaName

    @classmethod
    def makeMelaIndexes(cls) -> [int]:
        # these subChange numbers start at 1, not 0. Change 1 = Root
        indexes = []
        indexes.insert(0, False)
        indexes.insert(1, 1126)
        indexes.insert(2, 1127)
        indexes.insert(3, 1128)
        indexes.insert(4, 1129)
        indexes.insert(5, 1130)
        indexes.insert(6, 1131)
        indexes.insert(7, 1196)
        indexes.insert(8, 1197)
        indexes.insert(9, 1198)
        indexes.insert(10, 1199)
        indexes.insert(11, 1200)
        indexes.insert(12, 1201)
        indexes.insert(13, 1231)
        indexes.insert(14, 1232)
        indexes.insert(15, 1233)
        indexes.insert(16, 1234)
        indexes.insert(17, 1235)
        indexes.insert(18, 1236)
        indexes.insert(19, 1322)
        indexes.insert(20, 1323)
        indexes.insert(21, 1324)
        indexes.insert(22, 1325)
        indexes.insert(23, 1326)
        indexes.insert(24, 1327)
        indexes.insert(25, 1357)
        indexes.insert(26, 1358)
        indexes.insert(27, 1359)
        indexes.insert(28, 1360)
        indexes.insert(29, 1361)
        indexes.insert(30, 1362)
        indexes.insert(31, 1413)
        indexes.insert(32, 1414)
        indexes.insert(33, 1415)
        indexes.insert(34, 1416)
        indexes.insert(35, 1417)
        indexes.insert(36, 1418)
        indexes.insert(37, 1136)
        indexes.insert(38, 1137)
        indexes.insert(39, 1138)
        indexes.insert(40, 1139)
        indexes.insert(41, 1140)
        indexes.insert(42, 1141)
        indexes.insert(43, 1206)
        indexes.insert(44, 1207)
        indexes.insert(45, 1208)
        indexes.insert(46, 1209)
        indexes.insert(47, 1210)
        indexes.insert(48, 1211)
        indexes.insert(49, 1241)
        indexes.insert(50, 1242)
        indexes.insert(51, 1243)
        indexes.insert(52, 1244)
        indexes.insert(53, 1245)
        indexes.insert(54, 1246)
        indexes.insert(55, 1332)
        indexes.insert(56, 1333)
        indexes.insert(57, 1334)
        indexes.insert(58, 1335)
        indexes.insert(59, 1336)
        indexes.insert(60, 1337)
        indexes.insert(61, 1367)
        indexes.insert(62, 1368)
        indexes.insert(63, 1369)
        indexes.insert(64, 1370)
        indexes.insert(65, 1371)
        indexes.insert(66, 1372)
        indexes.insert(67, 1423)
        indexes.insert(68, 1424)
        indexes.insert(69, 1425)
        indexes.insert(70, 1426)
        indexes.insert(71, 1427)
        indexes.insert(72, 1428)
        return indexes

    @classmethod
    def getMelaIndexesFromScaleNames(cls, showDebug=False) -> [int]:
        # This wasn't working so fuck it
        _melaIndexes = []
        for m, mela in enumerate(cls.melaNameByMelaNumber[1:]):
            print(mela)
            _melaFound = False
            for i in range(2048):
                melaOptions = mela.split(",")
                for melaOption in melaOptions:
                    if melaOption in ScaleNames.namesBySequenceIndex[i]:
                        _melaFound = True
                        cls.pageFromMelaNumber.append(i)
                        break
                    else:
                        pass
                if _melaFound == True:
                    _melaIndexes.append(i)
                    Raga.pageFromMelaNumber.append(i)
            if not _melaFound:
                # print(Change.makeFromCarnaticNotation(input('Enter the carnatic notation: ')))
                raise TypeError(
                    mela + " not found in ScaleNames.namesBySequenceIndex\n"
                )
        if showDebug:
            print("here is Raga.pageFromMelaNumber: {}".format(Raga.pageFromMelaNumber))
        return _melaIndexes
Raga.pageFromMelaNumber = Raga.makeMelaIndexes()


class Braille:
    """Originally, I entered these in from top to bottom, meaning that the root position st was at the top left, and the fourth interval was bottom right."""

    ###Braille⠀
    # ⠁⠂⠃⠄⠅⠆⠇⠈⠉⠊⠋⠌⠍⠎⠏⠐⠑⠒⠓⠔⠕⠖⠗⠘⠙⠚⠛⠜⠝⠞⠟⠠⠡⠢⠣⠤⠥⠦⠧⠨⠩⠪⠫⠬⠭⠮⠯⠰⠱⠲⠳⠴⠵⠶⠷⠸⠹⠺⠻⠼⠽⠾⠿
    # ⡀⡁⡂⡃⡄⡅⡆⡇⡈⡉⡊⡋⡌⡍⡎⡏⡐⡑⡒⡓⡔⡕⡖⡗⡘⡙⡚⡛⡜⡝⡞⡟⡠⡡⡢⡣⡤⡥⡦⡧⡨⡩⡪⡫⡬⡭⡮⡯⡰⡱⡲⡳⡴⡵⡶⡷⡸⡹⡺⡻⡼⡽⡾⡿
    # ⢀⢁⢂⢃⢄⢅⢆⢇⢈⢉⢊⢋⢌⢍⢎⢏⢐⢑⢒⢓⢔⢕⢖⢗⢘⢙⢚⢛⢜⢝⢞⢟⢠⢡⢢⢣⢤⢥⢦⢧⢨⢩⢪⢫⢬⢭⢮⢯⢰⢱⢲⢳⢴⢵⢶⢷⢸⢹⢺⢻⢼⢽⢾⢿
    # ⣀⣁⣂⣃⣄⣅⣆⣇⣈⣉⣊⣋⣌⣍⣎⣏⣐⣑⣒⣓⣔⣕⣖⣗⣘⣙⣚⣛⣜⣝⣞⣟⣠⣡⣢⣣⣤⣥⣦⣧⣨⣩⣪⣫⣬⣭⣮⣯⣰⣱⣲⣳⣴⣵⣶⣷⣸⣹⣺⣻⣼⣽⣾⣿'''
    ##
    # http: // www.brailleauthority.org / ueb / symbols_list.pdf

    @classmethod
    def convertToTrigramatic(self):
        keys = Braille.semitonesFF.keys()
        kinds = (
            Braille.semitonesFF,
            Braille.semitonesFT,
            Braille.semitonesTF,
            Braille.semitonesTT,
        )

        for kind in kinds:
            print("{} = {{}}".format(kind))

        for key in keys:
            # for key in ((0,),(1,),(2,),(3,),(4,),(5,)):
            newKey = []
            for st in key:
                if st == 0:
                    newKey.append(4)
                if st == 1:
                    newKey.append(2)
                if st == 2:
                    newKey.append(0)
                if st == 3:
                    newKey.append(5)
                if st == 4:
                    newKey.append(3)
                if st == 5:
                    newKey.append(1)
            newKey.sort()

            newKey = tuple(newKey)
            # input(str(key)  + str(newKey) +  Braille.semitonesFF[key]+Braille.semitonesFF[newKey])
            for kind in kinds:
                if kind == Braille.semitonesFF:
                    kindName = "semitonesFF"
                elif kind == Braille.semitonesFT:
                    kindName = "semitonesFT"
                elif kind == Braille.semitonesTF:
                    kindName = "semitonesTF"
                elif kind == Braille.semitonesTT:
                    kindName = "semitonesTT"

                # input(str(key) + ' ' + str(kind))

                if True or key == (1, 3, 5):
                    print(
                        "{}[{}] = '{}' #{}{} {}".format(
                            kindName,
                            key,
                            kind[newKey],
                            Change(noteset=key).getHexagram(["symbol"])[0],
                            Change(noteset=key).getHexagram(["name"])[0],
                            kind[key],
                            key,
                        )
                    )
        input("copy paste that into Braille class")

    semitonesFF = {}
    semitonesFT = {}
    semitonesTF = {}
    semitonesTT = {}
    semitonesFF[(0, 1, 2, 3, 4, 5)] = "⠿"  # ䷀Creative ⠿
    semitonesFT[(0, 1, 2, 3, 4, 5)] = "⢿"  # ䷀Creative ⢿
    semitonesTF[(0, 1, 2, 3, 4, 5)] = "⡿"  # ䷀Creative ⡿
    semitonesTT[(0, 1, 2, 3, 4, 5)] = "⣿"  # ䷀Creative ⣿
    semitonesFF[()] = " "  # ䷁Receptive
    semitonesFT[()] = "⢀"  # ䷁Receptive ⢀
    semitonesTF[()] = "⡀"  # ䷁Receptive ⡀
    semitonesTT[()] = "⣀"  # ䷁Receptive ⣀
    semitonesFF[(0, 4)] = "⠔"  # ䷂Sprouting ⠅
    semitonesFT[(0, 4)] = "⢔"  # ䷂Sprouting ⢅
    semitonesTF[(0, 4)] = "⡔"  # ䷂Sprouting ⡅
    semitonesTT[(0, 4)] = "⣔"  # ䷂Sprouting ⣅
    semitonesFF[(1, 5)] = "⠊"  # ䷃Naïve ⠨
    semitonesFT[(1, 5)] = "⢊"  # ䷃Naïve ⢨
    semitonesTF[(1, 5)] = "⡊"  # ䷃Naïve ⡨
    semitonesTT[(1, 5)] = "⣊"  # ䷃Naïve ⣨
    semitonesFF[(0, 1, 2, 4)] = "⠗"  # ䷄Waiting ⠏
    semitonesFT[(0, 1, 2, 4)] = "⢗"  # ䷄Waiting ⢏
    semitonesTF[(0, 1, 2, 4)] = "⡗"  # ䷄Waiting ⡏
    semitonesTT[(0, 1, 2, 4)] = "⣗"  # ䷄Waiting ⣏
    semitonesFF[(1, 3, 4, 5)] = "⠺"  # ䷅Arguing ⠼
    semitonesFT[(1, 3, 4, 5)] = "⢺"  # ䷅Arguing ⢼
    semitonesTF[(1, 3, 4, 5)] = "⡺"  # ䷅Arguing ⡼
    semitonesTT[(1, 3, 4, 5)] = "⣺"  # ䷅Arguing ⣼
    semitonesFF[(1,)] = "⠂"  # ䷆Leading ⠈
    semitonesFT[(1,)] = "⢂"  # ䷆Leading ⢈
    semitonesTF[(1,)] = "⡂"  # ䷆Leading ⡈
    semitonesTT[(1,)] = "⣂"  # ䷆Leading ⣈
    semitonesFF[(4,)] = "⠐"  # ䷇Uniting ⠄
    semitonesFT[(4,)] = "⢐"  # ䷇Uniting ⢄
    semitonesTF[(4,)] = "⡐"  # ䷇Uniting ⡄
    semitonesTT[(4,)] = "⣐"  # ䷇Uniting ⣄
    semitonesFF[(0, 1, 2, 4, 5)] = "⠟"  # ䷈Homegrown ⠯
    semitonesFT[(0, 1, 2, 4, 5)] = "⢟"  # ䷈Homegrown ⢯
    semitonesTF[(0, 1, 2, 4, 5)] = "⡟"  # ䷈Homegrown ⡯
    semitonesTT[(0, 1, 2, 4, 5)] = "⣟"  # ䷈Homegrown ⣯
    semitonesFF[(0, 1, 3, 4, 5)] = "⠾"  # ䷉Treading ⠽
    semitonesFT[(0, 1, 3, 4, 5)] = "⢾"  # ䷉Treading ⢽
    semitonesTF[(0, 1, 3, 4, 5)] = "⡾"  # ䷉Treading ⡽
    semitonesTT[(0, 1, 3, 4, 5)] = "⣾"  # ䷉Treading ⣽
    semitonesFF[(0, 1, 2)] = "⠇"  # ䷊Peaceful ⠋
    semitonesFT[(0, 1, 2)] = "⢇"  # ䷊Peaceful ⢋
    semitonesTF[(0, 1, 2)] = "⡇"  # ䷊Peaceful ⡋
    semitonesTT[(0, 1, 2)] = "⣇"  # ䷊Peaceful ⣋
    semitonesFF[(3, 4, 5)] = "⠸"  # ䷋Obstructing ⠴
    semitonesFT[(3, 4, 5)] = "⢸"  # ䷋Obstructing ⢴
    semitonesTF[(3, 4, 5)] = "⡸"  # ䷋Obstructing ⡴
    semitonesTT[(3, 4, 5)] = "⣸"  # ䷋Obstructing ⣴
    semitonesFF[(0, 2, 3, 4, 5)] = "⠽"  # ䷌Fellow ⠷
    semitonesFT[(0, 2, 3, 4, 5)] = "⢽"  # ䷌Fellow ⢷
    semitonesTF[(0, 2, 3, 4, 5)] = "⡽"  # ䷌Fellow ⡷
    semitonesTT[(0, 2, 3, 4, 5)] = "⣽"  # ䷌Fellow ⣷
    semitonesFF[(0, 1, 2, 3, 5)] = "⠯"  # ䷍Prosperous ⠻
    semitonesFT[(0, 1, 2, 3, 5)] = "⢯"  # ䷍Prosperous ⢻
    semitonesTF[(0, 1, 2, 3, 5)] = "⡯"  # ䷍Prosperous ⡻
    semitonesTT[(0, 1, 2, 3, 5)] = "⣯"  # ䷍Prosperous ⣻
    semitonesFF[(2,)] = "⠁"  # ䷎Humble ⠂
    semitonesFT[(2,)] = "⢁"  # ䷎Humble ⢂
    semitonesTF[(2,)] = "⡁"  # ䷎Humble ⡂
    semitonesTT[(2,)] = "⣁"  # ䷎Humble ⣂
    semitonesFF[(3,)] = "⠠"  # ䷏Confident ⠐
    semitonesFT[(3,)] = "⢠"  # ䷏Confident ⢐
    semitonesTF[(3,)] = "⡠"  # ䷏Confident ⡐
    semitonesTT[(3,)] = "⣠"  # ䷏Confident ⣐
    semitonesFF[(0, 3, 4)] = "⠴"  # ䷐Following ⠕
    semitonesFT[(0, 3, 4)] = "⢴"  # ䷐Following ⢕
    semitonesTF[(0, 3, 4)] = "⡴"  # ䷐Following ⡕
    semitonesTT[(0, 3, 4)] = "⣴"  # ䷐Following ⣕
    semitonesFF[(1, 2, 5)] = "⠋"  # ䷑Correcting ⠪
    semitonesFT[(1, 2, 5)] = "⢋"  # ䷑Correcting ⢪
    semitonesTF[(1, 2, 5)] = "⡋"  # ䷑Correcting ⡪
    semitonesTT[(1, 2, 5)] = "⣋"  # ䷑Correcting ⣪
    semitonesFF[(0, 1)] = "⠆"  # ䷒Arriving ⠉
    semitonesFT[(0, 1)] = "⢃"  # ䷒Arriving ⢉
    semitonesTF[(0, 1)] = "⡃"  # ䷒Arriving ⡉
    semitonesTT[(0, 1)] = "⣃"  # ䷒Arriving ⣉
    semitonesFF[(4, 5)] = "⠘"  # ䷓Seeing ⠤
    semitonesFT[(4, 5)] = "⢘"  # ䷓Seeing ⢤
    semitonesTF[(4, 5)] = "⡘"  # ䷓Seeing ⡤
    semitonesTT[(4, 5)] = "⣘"  # ䷓Seeing ⣤
    semitonesFF[(0, 3, 5)] = "⠬"  # ䷔Biting ⠱
    semitonesFT[(0, 3, 5)] = "⢬"  # ䷔Biting ⢱
    semitonesTF[(0, 3, 5)] = "⡬"  # ䷔Biting ⡱
    semitonesTT[(0, 3, 5)] = "⣬"  # ䷔Biting ⣱
    semitonesFF[(0, 2, 5)] = "⠍"  # ䷕Adorning ⠣
    semitonesFT[(0, 2, 5)] = "⢍"  # ䷕Adorning ⢣
    semitonesTF[(0, 2, 5)] = "⡍"  # ䷕Adorning ⡣
    semitonesTT[(0, 2, 5)] = "⣍"  # ䷕Adorning ⣣
    semitonesFF[(5,)] = "⠈"  # ䷖Splitting ⠠
    semitonesFT[(5,)] = "⢈"  # ䷖Splitting ⢠
    semitonesTF[(5,)] = "⡈"  # ䷖Splitting ⡠
    semitonesTT[(5,)] = "⣈"  # ䷖Splitting ⣠
    semitonesFF[(0,)] = "⠄"  # ䷗Returning ⠁
    semitonesFT[(0,)] = "⢄"  # ䷗Returning ⢁
    semitonesTF[(0,)] = "⡄"  # ䷗Returning ⡁
    semitonesTT[(0,)] = "⣄"  # ䷗Returning ⣁
    semitonesFF[(0, 3, 4, 5)] = "⠼"  # ䷘Integrating ⠵
    semitonesFT[(0, 3, 4, 5)] = "⢼"  # ䷘Integrating ⢵
    semitonesTF[(0, 3, 4, 5)] = "⡼"  # ䷘Integrating ⡵
    semitonesTT[(0, 3, 4, 5)] = "⣼"  # ䷘Integrating ⣵
    semitonesFF[(0, 1, 2, 5)] = "⠏"  # ䷙Cultivating ⠫
    semitonesFT[(0, 1, 2, 5)] = "⢏"  # ䷙Cultivating ⢫
    semitonesTF[(0, 1, 2, 5)] = "⡏"  # ䷙Cultivating ⡫
    semitonesTT[(0, 1, 2, 5)] = "⣏"  # ䷙Cultivating ⣫
    semitonesFF[(0, 5)] = "⠌"  # ䷚Nourishing ⠡
    semitonesFT[(0, 5)] = "⢌"  # ䷚Nourishing ⢡
    semitonesTF[(0, 5)] = "⡌"  # ䷚Nourishing ⡡
    semitonesTT[(0, 5)] = "⣌"  # ䷚Nourishing ⣡
    semitonesFF[(1, 2, 3, 4)] = "⠳"  # ䷛Excessive ⠞
    semitonesFT[(1, 2, 3, 4)] = "⢳"  # ䷛Excessive ⢞
    semitonesTF[(1, 2, 3, 4)] = "⡳"  # ䷛Excessive ⡞
    semitonesTT[(1, 2, 3, 4)] = "⣳"  # ䷛Excessive ⣞
    semitonesFF[(1, 4)] = "⠒"  # ䷜Chasmic ⠌
    semitonesFT[(1, 4)] = "⢒"  # ䷜Chasmic ⢌
    semitonesTF[(1, 4)] = "⡒"  # ䷜Chasmic ⡌
    semitonesTT[(1, 4)] = "⣒"  # ䷜Chasmic ⣌
    semitonesFF[(0, 2, 3, 5)] = "⠭"  # ䷝Distanced ⠳
    semitonesFT[(0, 2, 3, 5)] = "⢭"  # ䷝Distanced ⢳
    semitonesTF[(0, 2, 3, 5)] = "⡭"  # ䷝Distanced ⡳
    semitonesTT[(0, 2, 3, 5)] = "⣭"  # ䷝Distanced ⣳
    semitonesFF[(2, 3, 4)] = "⠱"  # ䷞Magnetic ⠖
    semitonesFT[(2, 3, 4)] = "⢱"  # ䷞Magnetic ⢖
    semitonesTF[(2, 3, 4)] = "⡱"  # ䷞Magnetic ⡖
    semitonesTT[(2, 3, 4)] = "⣱"  # ䷞Magnetic ⣖
    semitonesFF[(1, 2, 3)] = "⠣"  # ䷟Persevering ⠚
    semitonesFT[(1, 2, 3)] = "⢣"  # ䷟Persevering ⢚
    semitonesTF[(1, 2, 3)] = "⡣"  # ䷟Persevering ⡚
    semitonesTT[(1, 2, 3)] = "⣣"  # ䷟Persevering ⣚
    semitonesFF[(2, 3, 4, 5)] = "⠹"  # ䷠Retiring ⠶
    semitonesFT[(2, 3, 4, 5)] = "⢹"  # ䷠Retiring ⢶
    semitonesTF[(2, 3, 4, 5)] = "⡹"  # ䷠Retiring ⡶
    semitonesTT[(2, 3, 4, 5)] = "⣹"  # ䷠Retiring ⣶
    semitonesFF[(0, 1, 2, 3)] = "⠧"  # ䷡Mighty ⠛
    semitonesFT[(0, 1, 2, 3)] = "⢧"  # ䷡Mighty ⢛
    semitonesTF[(0, 1, 2, 3)] = "⡧"  # ䷡Mighty ⡛
    semitonesTT[(0, 1, 2, 3)] = "⣧"  # ䷡Mighty ⣛
    semitonesFF[(3, 5)] = "⠨"  # ䷢Promoting ⠰
    semitonesFT[(3, 5)] = "⢨"  # ䷢Promoting ⢰
    semitonesTF[(3, 5)] = "⡨"  # ䷢Promoting ⡰
    semitonesTT[(3, 5)] = "⣨"  # ䷢Promoting ⣰
    semitonesFF[(0, 2)] = "⠅"  # ䷣Eclipsing ⠃
    semitonesFT[(0, 2)] = "⢅"  # ䷣Eclipsing ⢃
    semitonesTF[(0, 2)] = "⡅"  # ䷣Eclipsing ⡃
    semitonesTT[(0, 2)] = "⣅"  # ䷣Eclipsing ⣃
    semitonesFF[(0, 2, 4, 5)] = "⠝"  # ䷤Kindred ⠧
    semitonesFT[(0, 2, 4, 5)] = "⢝"  # ䷤Kindred ⢧
    semitonesTF[(0, 2, 4, 5)] = "⡝"  # ䷤Kindred ⡧
    semitonesTT[(0, 2, 4, 5)] = "⣝"  # ䷤Kindred ⣧
    semitonesFF[(0, 1, 3, 5)] = "⠮"  # ䷥Polarising ⠹
    semitonesFT[(0, 1, 3, 5)] = "⢮"  # ䷥Polarising ⢹
    semitonesTF[(0, 1, 3, 5)] = "⡮"  # ䷥Polarising ⡹
    semitonesTT[(0, 1, 3, 5)] = "⣮"  # ䷥Polarising ⣹
    semitonesFF[(2, 4)] = "⠑"  # ䷦Limping ⠆
    semitonesFT[(2, 4)] = "⢑"  # ䷦Limping ⢃
    semitonesTF[(2, 4)] = "⡑"  # ䷦Limping ⡃
    semitonesTT[(2, 4)] = "⣑"  # ䷦Limping ⣃
    semitonesFF[(1, 3)] = "⠢"  # ䷧Delivered ⠘
    semitonesFT[(1, 3)] = "⢢"  # ䷧Delivered ⢘
    semitonesTF[(1, 3)] = "⡢"  # ䷧Delivered ⡘
    semitonesTT[(1, 3)] = "⣢"  # ䷧Delivered ⣘
    semitonesFF[(0, 1, 5)] = "⠎"  # ䷨Declining ⠩
    semitonesFT[(0, 1, 5)] = "⢎"  # ䷨Declining ⢩
    semitonesTF[(0, 1, 5)] = "⡎"  # ䷨Declining ⡩
    semitonesTT[(0, 1, 5)] = "⣎"  # ䷨Declining ⣩
    semitonesFF[(0, 4, 5)] = "⠜"  # ䷩Benefitting ⠥
    semitonesFT[(0, 4, 5)] = "⢜"  # ䷩Benefitting ⢥
    semitonesTF[(0, 4, 5)] = "⡜"  # ䷩Benefitting ⡥
    semitonesTT[(0, 4, 5)] = "⣜"  # ䷩Benefitting ⣥
    semitonesFF[(0, 1, 2, 3, 4)] = "⠷"  # ䷪Deciding ⠟
    semitonesFT[(0, 1, 2, 3, 4)] = "⢷"  # ䷪Deciding ⢟
    semitonesTF[(0, 1, 2, 3, 4)] = "⡷"  # ䷪Deciding ⡟
    semitonesTT[(0, 1, 2, 3, 4)] = "⣷"  # ䷪Deciding ⣟
    semitonesFF[(1, 2, 3, 4, 5)] = "⠻"  # ䷫Coupling ⠾
    semitonesFT[(1, 2, 3, 4, 5)] = "⢻"  # ䷫Coupling ⢾
    semitonesTF[(1, 2, 3, 4, 5)] = "⡻"  # ䷫Coupling ⡾
    semitonesTT[(1, 2, 3, 4, 5)] = "⣻"  # ䷫Coupling ⣾
    semitonesFF[(3, 4)] = "⠰"  # ䷬Clustering ⠔
    semitonesFT[(3, 4)] = "⢰"  # ䷬Clustering ⢔
    semitonesTF[(3, 4)] = "⡰"  # ䷬Clustering ⡔
    semitonesTT[(3, 4)] = "⣰"  # ䷬Clustering ⣔
    semitonesFF[(1, 2)] = "⠃"  # ䷭Ascending ⠊
    semitonesFT[(1, 2)] = "⢃"  # ䷭Ascending ⢊
    semitonesTF[(1, 2)] = "⡃"  # ䷭Ascending ⡊
    semitonesTT[(1, 2)] = "⣃"  # ䷭Ascending ⣊
    semitonesFF[(1, 3, 4)] = "⠲"  # ䷮Confining ⠜
    semitonesFT[(1, 3, 4)] = "⢲"  # ䷮Confining ⢜
    semitonesTF[(1, 3, 4)] = "⡲"  # ䷮Confining ⡜
    semitonesTT[(1, 3, 4)] = "⣲"  # ䷮Confining ⣜
    semitonesFF[(1, 2, 4)] = "⠓"  # ䷯Sourced ⠎
    semitonesFT[(1, 2, 4)] = "⢓"  # ䷯Sourced ⢎
    semitonesTF[(1, 2, 4)] = "⡓"  # ䷯Sourced ⡎
    semitonesTT[(1, 2, 4)] = "⣓"  # ䷯Sourced ⣎
    semitonesFF[(0, 2, 3, 4)] = "⠵"  # ䷰Revolving ⠗
    semitonesFT[(0, 2, 3, 4)] = "⢵"  # ䷰Revolving ⢗
    semitonesTF[(0, 2, 3, 4)] = "⡵"  # ䷰Revolving ⡗
    semitonesTT[(0, 2, 3, 4)] = "⣵"  # ䷰Revolving ⣗
    semitonesFF[(1, 2, 3, 5)] = "⠫"  # ䷱Holding ⠺
    semitonesFT[(1, 2, 3, 5)] = "⢫"  # ䷱Holding ⢺
    semitonesTF[(1, 2, 3, 5)] = "⡫"  # ䷱Holding ⡺
    semitonesTT[(1, 2, 3, 5)] = "⣫"  # ䷱Holding ⣺
    semitonesFF[(0, 3)] = "⠤"  # ䷲Thundering ⠑
    semitonesFT[(0, 3)] = "⢤"  # ䷲Thundering ⢑
    semitonesTF[(0, 3)] = "⡤"  # ䷲Thundering ⡑
    semitonesTT[(0, 3)] = "⣤"  # ䷲Thundering ⣑
    semitonesFF[(2, 5)] = "⠉"  # ䷳Still ⠢
    semitonesFT[(2, 5)] = "⢉"  # ䷳Still ⢢
    semitonesTF[(2, 5)] = "⡉"  # ䷳Still ⡢
    semitonesTT[(2, 5)] = "⣉"  # ䷳Still ⣢
    semitonesFF[(2, 4, 5)] = "⠙"  # ䷴Infiltrating ⠦
    semitonesFT[(2, 4, 5)] = "⢙"  # ䷴Infiltrating ⢦
    semitonesTF[(2, 4, 5)] = "⡙"  # ䷴Infiltrating ⡦
    semitonesTT[(2, 4, 5)] = "⣙"  # ䷴Infiltrating ⣦
    semitonesFF[(0, 1, 3)] = "⠦"  # ䷵Betrothing ⠙
    semitonesFT[(0, 1, 3)] = "⢦"  # ䷵Betrothing ⢙
    semitonesTF[(0, 1, 3)] = "⡦"  # ䷵Betrothing ⡙
    semitonesTT[(0, 1, 3)] = "⣦"  # ䷵Betrothing ⣙
    semitonesFF[(0, 2, 3)] = "⠥"  # ䷶Abundant ⠓
    semitonesFT[(0, 2, 3)] = "⢥"  # ䷶Abundant ⢓
    semitonesTF[(0, 2, 3)] = "⡥"  # ䷶Abundant ⡓
    semitonesTT[(0, 2, 3)] = "⣥"  # ䷶Abundant ⣓
    semitonesFF[(2, 3, 5)] = "⠩"  # ䷷Wandering ⠲
    semitonesFT[(2, 3, 5)] = "⢩"  # ䷷Wandering ⢲
    semitonesTF[(2, 3, 5)] = "⡩"  # ䷷Wandering ⡲
    semitonesTT[(2, 3, 5)] = "⣩"  # ䷷Wandering ⣲
    semitonesFF[(1, 2, 4, 5)] = "⠛"  # ䷸Osmotic ⠮
    semitonesFT[(1, 2, 4, 5)] = "⢛"  # ䷸Osmotic ⢮
    semitonesTF[(1, 2, 4, 5)] = "⡛"  # ䷸Osmotic ⡮
    semitonesTT[(1, 2, 4, 5)] = "⣛"  # ䷸Osmotic ⣮
    semitonesFF[(0, 1, 3, 4)] = "⠶"  # ䷹Joyous ⠝
    semitonesFT[(0, 1, 3, 4)] = "⢶"  # ䷹Joyous ⢝
    semitonesTF[(0, 1, 3, 4)] = "⡶"  # ䷹Joyous ⡝
    semitonesTT[(0, 1, 3, 4)] = "⣶"  # ䷹Joyous ⣝
    semitonesFF[(1, 4, 5)] = "⠚"  # ䷺Dispersing ⠬
    semitonesFT[(1, 4, 5)] = "⢚"  # ䷺Dispersing ⢬
    semitonesTF[(1, 4, 5)] = "⡚"  # ䷺Dispersing ⡬
    semitonesTT[(1, 4, 5)] = "⣚"  # ䷺Dispersing ⣬
    semitonesFF[(0, 1, 4)] = "⠖"  # ䷻Composed ⠍
    semitonesFT[(0, 1, 4)] = "⢖"  # ䷻Composed ⢍
    semitonesTF[(0, 1, 4)] = "⡖"  # ䷻Composed ⡍
    semitonesTT[(0, 1, 4)] = "⣖"  # ䷻Composed ⣍
    semitonesFF[(0, 1, 4, 5)] = "⠞"  # ䷼True ⠭
    semitonesFT[(0, 1, 4, 5)] = "⢞"  # ䷼True ⢭
    semitonesTF[(0, 1, 4, 5)] = "⡞"  # ䷼True ⡭
    semitonesTT[(0, 1, 4, 5)] = "⣞"  # ䷼True ⣭
    semitonesFF[(2, 3)] = "⠡"  # ䷽Detailed ⠒
    semitonesFT[(2, 3)] = "⢡"  # ䷽Detailed ⢒
    semitonesTF[(2, 3)] = "⡡"  # ䷽Detailed ⡒
    semitonesTT[(2, 3)] = "⣡"  # ䷽Detailed ⣒
    semitonesFF[(0, 2, 4)] = "⠕"  # ䷾Completed ⠇
    semitonesFT[(0, 2, 4)] = "⢕"  # ䷾Completed ⢇
    semitonesTF[(0, 2, 4)] = "⡕"  # ䷾Completed ⡇
    semitonesTT[(0, 2, 4)] = "⣕"  # ䷾Completed ⣇
    semitonesFF[(1, 3, 5)] = "⠪"  # ䷿Incomplete ⠸
    semitonesFT[(1, 3, 5)] = "⢪"  # ䷿Incomplete ⢸
    semitonesTF[(1, 3, 5)] = "⡪"  # ䷿Incomplete ⡸
    semitonesTT[(1, 3, 5)] = "⣪"  # ䷿Incomplete ⣸


from __future__ import annotations, print_function
import colorsys
from Utility import Utility
input = Utility.input
class Colour:
    A4 = 432
    # These colours will be the default, but then it will replace it after.
    makeGraphicsColoured = 0
    useBlackBackground = 0
    """rgbByDistanceColourDk = ['008700', '8eff29', 'ffff0d', 'ff9b00',
                        'ff6600', 'c02e29', 'ff0000', 'dc00b9',
                        '8507b9', '6007ff', '0000ff', '00ae62']

    rgbByDistanceColourLt = ['008700', '8eff29', 'ffff0d', 'ff9b00',
                        'ff6600', 'c02e29', 'ff0000', 'dc00b9',
                        '8507b9', '6007ff', '0000ff', '00ae62']"""

    if False:
        hues = [i / 12 for i in range(12)]
        # Transpose from start == red to start == green
        hues = hues[5:] + hues[:5]

    else:
        hues = [
            0,  # Red
            1.0 / 24,  # Red-Orange
            2.0 / 24,  # Orange
            3.1 / 24,  # Honey Eb
            4 / 24,  # Yellow
            5.2 / 24,  # Chartreuse
            7 / 24,  # Green
            10 / 24,
            13 / 24,  # Blue
            17.5 / 24,  # Blurple
            19.1 / 24,  # Purple
            21 / 24,
        ]
        """        hues = [0, 2 / 24, 2 / 24,
                3 / 24, 4 / 6, 2 / 9,
                7 / 24, 10 / 24, 13 / 24,
                18 / 24, 19.5 / 24, 21 / 24]"""
        hues = hues[7:] + hues[:7]
    hues = hues[::-1]
    hlsByDistanceColourDk = [(hue, 0.45, 1) for hue in hues]  # .2
    hlsByDistanceColourLt = [(hue, 0.67, 1) for hue in hues]
    hlsByDistanceColourNt = [(hue, 0.5, 1) for hue in hues]
    # input(hlsByDistanceColourDk)
    rgbByDistanceColourDk = [
        colorsys.hls_to_rgb(hls[0], hls[1], hls[2]) for hls in hlsByDistanceColourDk
    ]
    rgbByDistanceColourLt = [
        colorsys.hls_to_rgb(hls[0], hls[1], hls[2]) for hls in hlsByDistanceColourLt
    ]
    rgbByDistanceColourNt = [
        colorsys.hls_to_rgb(hls[0], hls[1], hls[2]) for hls in hlsByDistanceColourNt
    ]

    rgbByDistanceColourDk = [
        "".join("%02X" % round(i * 255) for i in rgb) for rgb in rgbByDistanceColourDk
    ]
    rgbByDistanceColourLt = [
        "".join("%02X" % round(i * 255) for i in rgb) for rgb in rgbByDistanceColourLt
    ]
    rgbByDistanceColourNt = [
        "".join("%02X" % round(i * 255) for i in rgb) for rgb in rgbByDistanceColourNt
    ]

    rgbInkColourDk = "FFFFFF"
    rgbInkColourLt = "000000"
    rgbInkColourNt = "FFFFFF"
    rgbPaperColourDk = "000000"
    rgbPaperColourLt = "FFFFFF"
    rgbPaperColourNt = "000000"
    # input('rgbByDistanceColourDk {}\nrgbByDistanceColourLt {}'.format(rgbByDistanceColourDk,rgbByDistanceColourLt))
    rgbByDistanceBWDk = ["CCCCCC"] * 12
    rgbByDistanceBWLt = ["444444"] * 12
    rgbByDistanceBWNt = ["888888"] * 12
    if makeGraphicsColoured == False:
        rgbByDistance = [i for i in rgbByDistanceBWDk]
    else:
        rgbByDistance = [i for i in rgbByDistanceColourDk]
    # nameByDistance = ['1','b2','2','b3','3','4','b5','5','b6','6','b7','7']
    nameByDistanceLt = [
        "C-Lt",
        "Db-Lt",
        "D-Lt",
        "Eb-Lt",
        "E-Lt",
        "F-Lt",
        "Gb-Lt",
        "G-Lt",
        "Ab-Lt",
        "A-Lt",
        "Bb-Lt",
        "B-Lt",
    ]
    nameByDistanceDk = [
        "C-Dk",
        "Db-Dk",
        "D-Dk",
        "Eb-Dk",
        "E-Dk",
        "F-Dk",
        "Gb-Dk",
        "G-Dk",
        "Ab-Dk",
        "A-Dk",
        "Bb-Dk",
        "B-Dk",
    ]
    nameByDistanceNt = ["C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B"]
    validLatexTags = nameByDistanceLt + nameByDistanceDk + nameByDistanceNt
    # nameByDistance = ['green','green-yellow','yellow','yellow-orange','orange','orange-red','red','red-purple','purple','purple-blue','blue']

    def __init__(self, A4=432, makeGraphicsColoured=True):
        self.A4 = A4
        self.makeGraphicsColoured = makeGraphicsColoured
        # These colours will be the default, but then it will replace it after.
        self.rgbByDistanceColourDk = [
            "008700",
            "8eff29",
            "ffff0d",
            "ff9b00",
            "ff6600",
            "c02e29",
            "ff0000",
            "dc00b9",
            "8507b9",
            "6007ff",
            "0000ff",
            "00ae62",
        ]
        self.rgbByDistanceBWDk = [
            "000000",
            "000000",
            "000000",
            "000000",
            "000000",
            "000000",
            "000000",
            "000000",
            "000000",
        ]
        if makeGraphicsColoured == False:
            self.rgbByDistance = self.rgbByDistanceBWDk
        else:
            self.rgbByDistance = self.rgbByDistanceColourDk

    @classmethod
    def blendColourValue(cls, a, b, t):
        """
        :param a: first colour value
        :param b: second colour value
        :param t: transition between
        :return average value:
        """
        return math.sqrt((1 - t) * a**2 + t * b**2)

    @classmethod
    def blendAlphaValue(cls, a, b, t):
        return (1 - t) * a + t * b

    @classmethod
    def blendColours(cls, c1, c2, t):
        """
        :param c1: colour 1
        :param c2: colour 2
        :param t: transition
        :return:
        """
        usingAlpha = len(c1) in (2, 4) == True
        ret = []
        for b1, band1 in enumerate(c1):
            band2 = c2[b1]
            if b1 != len(c1) or not usingAlpha:
                raise TypeError("not done")

    @classmethod
    def writeColourSchemesToFile(cls, fileFormat="tex"):

        _jsStr = ""
        _schemes = ("Colour", "ColourInv", "BW", "BWInv")
        for s, scheme in enumerate(_schemes):
            print("generating", scheme)
            if fileFormat == "tex":
                _filename = os.path.join(Project.directoryTex , "colourMap-" + scheme + ".tex")
            elif fileFormat == "js":
                _filename = os.path.join(Project.directoryJS, "colourMap.js")
            else:
                raise ValueError("non valid format")
            if scheme == "Colour":
                _colours = Colour.rgbByDistanceColourDk
                _coloursInv = Colour.rgbByDistanceColourLt
            elif scheme == "ColourInv":
                _colours = Colour.rgbByDistanceColourLt
                _coloursInv = Colour.rgbByDistanceColourDk
            elif scheme == "BW":
                _colours = ["000000"] * 12
                _coloursNeut = ["000000"] * 12
                _coloursInv = ["FFFFFF"] * 12
            elif scheme == "BWInv":
                _colours = ["FFFFFF"] * 12
                _coloursInv = ["000000"] * 12
                _coloursNeut = ["FFFFFF"] * 12
            if scheme in ("Colour", "ColourInv"):
                _coloursNeut = Colour.rgbByDistanceColourNt

            _str = ""

            if fileFormat == "js" and s == 0:
                _str += (
                    "const keyColours = {\n"
                    + "".join(['  "' + i + '": {},\n' for i in _schemes])
                    + "};\n"
                )
            for p, palette in enumerate((_colours, _coloursNeut, _coloursInv)):

                if p == 0:
                    colourTags = Colour.nameByDistanceDk
                    inkColour = Colour.rgbInkColourDk
                    paperColour = Colour.rgbPaperColourDk
                elif p == 1:
                    colourTags = Colour.nameByDistanceNt
                    inkColour = Colour.rgbInkColourNt
                    paperColour = Colour.rgbPaperColourNt
                elif p == 2:
                    colourTags = Colour.nameByDistanceLt
                    inkColour = Colour.rgbInkColourLt
                    paperColour = Colour.rgbPaperColourLt
                for c, colour in enumerate(palette):
                    print(colourTags[c], colour)
                    if fileFormat == "tex":
                        _str += (
                            "\definecolor{"
                            + colourTags[c]
                            + "}{HTML}{"
                            + colour
                            + "}%\n"
                        )
                    elif fileFormat == "js":
                        _str += 'keyColours["{}"]["{}"] = "#{}";\n'.format(
                            scheme, colourTags[c], colour
                        )
                if fileFormat == "tex":
                    _str += "\definecolor{ink}{HTML}{" + inkColour + "}%\n"
                    _str += "\definecolor{paper}{HTML}{" + paperColour + "}%\n"
                elif fileFormat == "js":

                    if "Dk" in colourTags[0]:
                        darkness = "-Dk"
                    elif "Lt" in colourTags[0]:
                        darkness = "-Lt"
                    elif "Nt" in colourTags[0]:
                        darkness = "-Nt"
                    else:
                        darkness = ""
                    _str += 'keyColours["{}"]["{}"] = "#{}";\n'.format(
                        scheme, "ink" + darkness, inkColour
                    )
                    _str += 'keyColours["{}"]["{}"] = "#{}";\n'.format(
                        scheme, "paper" + darkness, paperColour
                    )

                # _str += '\n'

            if _str[-1] == "\n":
                _str = _str[0:-1]
            text_file = open(_filename, "w+")
            text_file.write(_str)
            text_file.close()
            _jsStr += _str

            # print(_str)

            print("updated colourMap at " + _filename)
            # input(_str + '\nwould thou like to continue?')
        if fileFormat == "js":
            text_file = open(_filename, "w+")
            text_file.write(_jsStr)
            text_file.close()

    @classmethod
    def getGlobalRGBsFromA4(cls, f: int) -> []:
        # not finished
        raise TypeError("not done")
        # using well tempered tuning
        if f in (None, 0):
            f = cls.A4
        RGBs = []
        freqs = []
        # grab the frequency of C, 9 notes below
        # because the book assumes that the 'first' place
        # is C, not A
        C4 = f * 2 ** (-9 / 12)
        for i in range(12):
            freqs.append(C4 * 2 ** (i / 12))
        # TODO: calculate hue
        # Then get rgb for that
        # Then

    @classmethod
    def getTransposedColourTags(
        cls,
        colourTranspose: int = 0,
        makeGreyscale=False,
        invertColour=False,
        pianoKeys=False,
        neutralColours=False,
    ):
        if makeGreyscale:
            if not pianoKeys:
                if invertColour:
                    _colours = Colour.nameByDistanceLt
                elif not invertColour:
                    _colours = Colour.nameByDistanceDk
                if neutralColours:
                    _colours = Colour.nameByDistanceNt
            elif pianoKeys:
                _colours = [0 for i in range(12)]
                for i in range(len(_colours)):
                    if i in (0, 2, 4, 5, 7, 9, 11):
                        if invertColour:
                            _colours[i] = "black"
                        elif not invertColour:
                            _colours[i] = "white"
                    else:
                        if invertColour:
                            _colours[i] = "white"
                        elif not invertColour:
                            _colours[i] = "black"
        elif not makeGreyscale:
            if invertColour:
                _colours = Colour.nameByDistanceLt
            elif not invertColour:
                _colours = Colour.nameByDistanceDk
            if neutralColours:
                _colours = Colour.nameByDistanceNt

        if len(_colours) != 12:
            raise (
                TypeError(
                    "There are supposed to be 12 colours but there are not. length {}\n{} ".format(
                        len(_colours), _colours
                    )
                )
            )
        return [_colours[(i + colourTranspose) % 12] for i in range(12)]

    @classmethod
    def getTransposedColours(
        cls,
        colourTranspose: int = 0,
        greyScale=False,
        invertColour=False,
        pianoKeys=False,
        neutralColours=False,
    ):
        if greyScale:
            if not pianoKeys:
                if neutralColours:
                    _colours = Colour.rgbaTupleByDistanceBWNt
                elif invertColour:
                    _colours = Colour.rgbaTupleByDistanceBWLt
                elif not invertColour:
                    _colours = Colour.rgbaTupleByDistanceBWDk
            elif pianoKeys:
                _colours = [0 for i in range(12)]
                for i in range(len(_colours)):
                    if i in (0, 2, 4, 5, 7, 9, 11):
                        if invertColour:
                            _colours[i] = (0, 0, 0, 255)
                        elif not invertColour:
                            _colours[i] = (255, 255, 255, 255)
                    else:
                        if invertColour:
                            _colours[i] = (255, 255, 255, 255)
                        elif not invertColour:
                            _colours[i] = (0, 0, 0, 255)
        elif not greyScale:
            if neutralColours:
                _colours = Colour.rgbaTupleByDistanceColourNt
            elif invertColour:
                _colours = Colour.rgbaTupleByDistanceColourLt
            elif not invertColour:
                _colours = Colour.rgbaTupleByDistanceColourDk
        if len(_colours) != 12:
            raise (
                TypeError(
                    "There are supposed to be 12 colours but there are not. length {}\n{} ".format(
                        len(_colours), _colours
                    )
                )
            )
        return [_colours[(i + colourTranspose) % 12] for i in range(12)]

    @classmethod
    def getColourOfChange(cls, change: Change):
        raise TypeError("not done")

    @classmethod
    def getRGBTuples(cls, includeAlpha=False, alpha=255) -> list[tuple]:
        _tuples = []
        for i in cls.rgbByDistance:
            _tuples.append(cls.convertRGBToTupleForm(i))
        return _tuples

    @classmethod
    def convertRGBToTupleForm(cls, rgb: str, includeAlpha=False, alpha=255) -> tuple:

        first = rgb[:2]
        second = rgb[2:4]
        third = rgb[4:]
        if includeAlpha == False:
            return tuple([int(first, 16), int(second, 16), int(third, 16)])
        else:
            return tuple([int(first, 16), int(second, 16), int(third, 16), alpha])

    @classmethod
    def modifyAlphaChannelOfRGBATuple(cls, rgba: tuple, alpha: int) -> tuple:
        if not len(rgba) == 4:
            raise TypeError("rgba tuple supposed to have 4 channels." + str(rgba))
        if alpha < 0 or alpha > 255 or type(alpha) != int:
            raise ValueError("alpha to be supplied as int between 0-255. " + str(alpha))
        return tuple([rgba[0], rgba[1], rgba[2], alpha])
        
# Finish initialising
Colour.rgbaTupleByDistance = [
    Colour.convertRGBToTupleForm(i, includeAlpha=True) for i in Colour.rgbByDistance
]
Colour.rgbTupleByDistance = [
    Colour.convertRGBToTupleForm(i, includeAlpha=False) for i in Colour.rgbByDistance
]
Colour.rgbaTupleByDistanceColourDk = [
    Colour.convertRGBToTupleForm(i, includeAlpha=True)
    for i in Colour.rgbByDistanceColourDk
]
Colour.rgbTupleByDistanceColourDk = [
    Colour.convertRGBToTupleForm(i, includeAlpha=False)
    for i in Colour.rgbByDistanceColourDk
]
Colour.rgbaTupleByDistanceColourNt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=True)
    for i in Colour.rgbByDistanceColourNt
]
Colour.rgbTupleByDistanceColourNt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=False)
    for i in Colour.rgbByDistanceColourNt
]
Colour.rgbaTupleByDistanceColourLt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=True)
    for i in Colour.rgbByDistanceColourLt
]
Colour.rgbTupleByDistanceColourLt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=False)
    for i in Colour.rgbByDistanceColourLt
]
Colour.rgbaTupleByDistanceBWLt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=True) for i in Colour.rgbByDistanceBWLt
]
Colour.rgbTupleByDistanceBWLt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=False)
    for i in Colour.rgbByDistanceBWLt
]
Colour.rgbaTupleByDistanceBWDk = [
    Colour.convertRGBToTupleForm(i, includeAlpha=True) for i in Colour.rgbByDistanceBWDk
]
Colour.rgbTupleByDistanceBWDk = [
    Colour.convertRGBToTupleForm(i, includeAlpha=False)
    for i in Colour.rgbByDistanceBWDk
]
Colour.rgbaTupleByDistanceBWNt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=True) for i in Colour.rgbByDistanceBWNt
]
Colour.rgbTupleByDistanceBWNt = [
    Colour.convertRGBToTupleForm(i, includeAlpha=False)
    for i in Colour.rgbByDistanceBWNt
]


class Zodiac:
    # Astrological signs: ♈     ♉      ♊      ♋      ♌   ♍     ♎     ♏       ♐           ♑         ♒        ♓
    #                    Aries Taurus Gemini Cancer Leo Virgo Libra Scorpio Sagittarius Capricorn Aquarius Pisces
    #                    0     1      2      3      4   5     6     7       8           9         10       11
    semitoneToZodiac = ["♈", "♉", "♊", "♋", "♌", "♍", "♎", "♏", "♐", "♑", "♒", "♓"]
    semitoneToSign = [
        "Aries",
        "Taurus",
        "Gemini",
        "Cancer",
        "Leo",
        "Virgo",
        "Libra",
        "Scorpio",
        "Sagittarius",
        "Capricorn",
        "Aquarius",
        "Pisces",
    ]
    semitoneToMonth = [
        "January",
        "Febuary",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ]


from __future__ import annotations, print_function
from Utility import Utility
from Change import Change
from JazzNote import JazzNote
from Book import Book
from Colour import Colour
from Latex import Latex
import numpy as np
from tqdm import tqdm
from PIL import Image, ImageDraw, ImageFont
import math,os,pdf2image

input = Utility.input
class FCircle:
    # white = (255, 255, 255, 255)
    canvasBGColour = (0, 0, 0, 0)
    circleBGColourInv = (255, 255, 255, 255)
    circleBGColour = (0, 0, 0, 255)
    lineColour = (0, 0, 0, 255)
    lineColourInv = (255, 255, 255, 255)  # (0,0,0,255)
    minimumCirclePixels = 2
    proportion = 0.618033988749894848204586834
    artCircleProportion = 0.618033988749894848204586834 * 2 ** (1 / 2)
    lineProportion = 0.04258684210526316
    radiusOuterProportionRCircle = proportion * 1.3
    radiusOfEachRCircle = 0.15
    singleNoteCircleProportion = proportion
    fractality = 3
    vectorSmallResolution = 64
    vectorMediumResolution = 256
    vectorLargeResolution = 2048
    vectorSaveResolutions = [
        vectorSmallResolution,
        vectorMediumResolution,
        vectorLargeResolution,
    ]
    polygonResolution = 64
    saveOnlyOneKeyIfResolutionExceeds = 2048
    validTypes = ["FCircle", "SCircle", "PCircle"]

    def __init__(
        self,
        centre: list[float, float],
        r: [float],
        change: Change,
        colourTranspose: int,
        proportion=None,
        invertBG=False,
    ):
        self.centre = centre
        self.x = centre[0]
        self.y = centre[1]
        if proportion is None:
            proportion = FCircle.proportion
        self.r = r * proportion
        #input( change.__class__.__name__)
        if change.__class__.__name__ == "Change":
            self.change = change
        else:
            raise TypeError(
                "change needs to be type Change, not "
                + str(type(change))
                + " "
                + str(change)
                + '. '
                + change.__class__.__name__
            )
        self.colourTranspose = colourTranspose
        self.invertBG = invertBG

    def __repr__(self):
        return "FCircle(centre={},r={},change=Change({}),colourTranspose={},invertBG={})".format(
            self.centre,
            self.r,
            self.change.__repr__(),
            self.colourTranspose,
            self.invertBG,
        )

    def __str__(self):
        return "FCircle({},{},{},{})".format(
            self.centre, self.r, self.change, self.colourTranspose
        )

    @classmethod
    def setClassConstantsForSCircle(cls):
        FCircle.includeRCircle = False
        FCircle.useChordCircles = True
        FCircle.proportion = 1
        FCircle.lineProportion = 0.0001
        FCircle.alternateBG = False

    @classmethod
    def setClassConstantsForFCircle(cls):
        FCircle.includeRCircle = True
        FCircle.useChordCircles = False
        FCircle.alternateBG = True
        FCircle.proportion = 0.618033988749894848204586834
        FCircle.lineProportion = 0.04258684210526316
        FCircle.alternateBG = True

    def drawTrigrams(self, im, colours, lineColour, thickness, showDebug=False):

        draw = ImageDraw.Draw(im)
        r2 = self.r * 2.0
        r = self.r / FCircle.proportion
        centreX = self.r
        centreY = self.r
        pi = math.pi
        lineProportion = 1 / 12.0
        # https://math.stackexchange.com/questions/377234/circle-in-square-calculate-distance-from-squares-corner-to-circles-perimeter
        # diagonal = 0.5 * ( r2 * 2 ** -1 - r2)
        diagonal = 0.5 * r2 * (2**-1 - 1)
        trigramH = diagonal / (2**0.5)  # 2#(2*pi)

        _trigrams = self.change.divideScaleBy(denominator=4, normaliseToSlice=True)
        _qturn = math.radians(90)
        _width = trigramH * 3 / 2.0
        st = 0
        for t, trigram in enumerate(_trigrams):
            # _angle = math.radians(90 * t + 45 + 270)
            _angle = math.radians(90 * t + 45 + 270)
            # input('angle {}'.format(math.degrees(_angle)))
            _lines = trigram.getSemitones()
            # first number is the radius of circle where trigrams start

            x = 1 * math.cos(_angle) * r + r
            y = 1 * math.sin(_angle) * r + r
            x += int(math.cos(_angle) * diagonal / pi)
            y += int(math.sin(_angle) * diagonal / pi)
            for l in range(3):

                x -= int(math.cos(_angle) * diagonal / pi)
                y -= int(math.sin(_angle) * diagonal / pi)
                _colour = colours[st % len(colours)]
                if l in _lines:
                    # Unbroken line
                    x1 = x + (math.cos(_angle - _qturn) * _width / 2)
                    y1 = y + (math.sin(_angle - _qturn) * _width / 2)
                    x2 = x + (math.cos(_angle + _qturn) * _width / 2)
                    y2 = y + (math.sin(_angle + _qturn) * _width / 2)

                    Graphics.fatLine(
                        im=im,
                        line=[x1, y1, x2, y2],
                        outline=lineColour,
                        fill=_colour,
                        thickness=thickness,
                    )

                else:
                    # broken line
                    Graphics.fatLine(
                        im=im,
                        line=[
                            x + (math.cos(_angle - _qturn) * _width / 2),
                            y + (math.sin(_angle - _qturn) * _width / 2),
                            x - (math.cos(_angle + _qturn) * _width / 12),
                            y - (math.sin(_angle + _qturn) * _width / 12),
                        ],
                        fill=_colour,
                        outline=lineColour,
                        thickness=thickness,
                    )
                    Graphics.fatLine(
                        im,
                        [
                            x - (math.cos(_angle - _qturn) * _width / 2),
                            y - (math.sin(_angle - _qturn) * _width / 2),
                            x + (math.cos(_angle + _qturn) * _width / 12),
                            y + (math.sin(_angle + _qturn) * _width / 12),
                        ],
                        fill=_colour,
                        outline=lineColour,
                        thickness=thickness,
                    )

                    """Graphics.fatLine(im, [x - (math.cos(_angle - _qturn) * _width / 2),
                                          y - (math.sin(_angle - _qturn) * _width / 2),
                                          x + (math.cos(_angle + _qturn) * _width / 12),
                                          y + (math.sin(_angle + _qturn) * _width / 12),
                                          ], fill=_colour, outline=lineColour, thickness=thickness
                                     )"""
                    print(st)
                st += 1

        if showDebug:
            im.show()
        return im

    @classmethod
    def renderFCircleGrid(
        cls,
        columns: int,
        rows: int,
        colourKey: str = None,
        useHexagramGrid=True,
        sortHexIntoSubsequence=False,
        mirrorNegativeAxis=True,
        splitX=4,
        splitY=4,
        greyScale=False,
        invertColour=False,
        skipIfFileExists=False,
        circleType='FCircle'
    ):

        #raise TypeError("shit")
        # _sPx = Latex.SmallCircleResolution
        if colourKey is None:
            colourKey = Book.rootColourKey

        if useHexagramGrid:
            if sortHexIntoSubsequence:
                _subname = "_hexagram-subsequence"
            else:
                _subname = "_hexagram-iChing"
        elif not useHexagramGrid:
            _subname = "_sequence"

        Utility.makeDirectory(os.path.join(Graphics.fCircleChartsDirectory, _subname))
        if not invertColour:
            if not greyScale:
                _directory = os.path.join(Graphics.fCircleChartsDirectory, _subname)
            else:
                _directory = os.path.join(Graphics.fCircleChartsBWDirectory, _subname)
        else:
            if not greyScale:
                _directory = os.path.join(Graphics.fCircleChartsInvDirectory, _subname)
            else:
                _directory = (
                    os.path.join(Graphics.fCircleChartsBWInvDirectory, _subname)
                )

        _sPx = 64
        _sep = round(_sPx / 3.14159**3)
        _xpad = 0
        _ypad = round(0 + (_sPx / 0.772727272727272) - _sPx)
        _cellWidth = _sPx + _sep + _xpad
        _cellHeight = _sPx + _sep + _ypad
        # input('imsgeSize {} circleSize {}'.format((columns*(_sPx+_sep)),_sPx))
        _paddingRows = 0
        _paddingRows = 0
        til = Image.new("RGB", (columns * _cellWidth + _sep, rows * _cellHeight + _sep))
        draw = ImageDraw.Draw(til)
        font = ImageFont.truetype(Latex.fontStr, round(17))
        # _fontHeight = font.getSize('M')[1]

        if sortHexIntoSubsequence:
            _hexagramBook = Book(["1", "b2", "2", "b3", "3", "4"])
            # input(str(_hexagramBook.sequence))
            _positiveHexagrams = [_hexagramBook[i] for i in range(0, 32, +1)]
            _negativeHexagrams = [i.withoutNote("1") for i in _positiveHexagrams]
            _negativeHexagrams.reverse()

            _hexaST = [
                i.getSemitones() for i in _negativeHexagrams + _positiveHexagrams
            ]

            # for i in _positiveHexagrams:
            # _hexaST.append(i.getSemitones())

            print(
                "positive",
                _positiveHexagrams,
                "negative",
                _negativeHexagrams,
                "length",
                len(_hexaST),
                "hexaST",
                _hexaST,
                sep="\n",
            )
            """_hexagramNumbers = ['adsfg' for i in range(65)]
            for i in range(1,65):
                _hexagramNumbers[i] = Hexagram.subChange[i]
                print('asdfasdfasdf',_hexagramNumbers[i])
            #input(_hexagramNumbers)#[Change.makeFromSet(s).getHexagram(['subpage']) for s in _hexaST]
            _hexaST = [Hexagram.notesets[_hexagramNumbers[i]] for i in _hexagramNumbers[1:]]"""
        elif not sortHexIntoSubsequence:
            _hexaST = [s for s in Hexagram.notesets[1:]]
        _changeST = _hexaST
        print("started drawing. ", end="")
        _cell = 0
        for i1, h1 in tqdm(enumerate(_hexaST)):
            _cell += 1
            print("row", i1)
            for i2, h2 in enumerate(_hexaST):
                _cell += 1
                # _changeNumber=Change.makeFromSet(h1+[i+6 for i in h2]).getChangeNumber(decorateChapter=False,addOneToBookPage=True)
                if useHexagramGrid:

                    _change = Change.makeFromSet(h1 + [i + 6 for i in h2])
                elif not useHexagramGrid:
                    _pageNumber = i1 * 64 + i2 - 2048
                    # print(_pageNumber)
                    # TODO: If you change sequence indexing change rootIsChangeOne
                    _change = Change.makeFromChangeNumber(
                        _pageNumber,
                    )  # i1*64+i2-2048)

                # print(h1,h2,_change)
                #### Do it this way if you want to grab the src imgs from bitmaps
                # im = Image.open(Graphics.getFCirclePath(
                #    _change, colourKey, _sPx, includeFilename=True))  # 25x25
                #### Do it this way if you want to grab the src imgs from pdfs



                im = pdf2image.convert_from_path(
                    Graphics.renderFCircle(
                        change=_change,
                        resolution=256,
                        rootColourKey=Book.rootColourKey,
                        invertColour=invertColour,
                        greyScale=greyScale,
                        saveOnlyOneKey=True,  # resolution >= 4096,
                        filetypes=['pdf'],
                        circleType=circleType,
                        skipIfFileExists=True,
                    ),
                    size=_sPx,
                )[0].convert("RGBA")

                _x = _cellWidth * i2 + _sep
                _y = _cellHeight * i1 + _sep
                _txt = (
                    "".join(_change.getHexagram())
                    + " " * (4 - len(_change))
                    + str(_change.getChangeNumber())
                )
                til.paste(im.convert("RGBA"), (_x, _y))
                draw.text(
                    (_x, _y + _sPx + _ypad / 3), _txt, font=font, fill=(255, 255, 255)
                )

        _filename = "FCircleChart"
        _extension = ".png"

        if splitX > 1 or splitY > 1:

            _cropDirectory = os.path.join(_directory, "crop{}by{}\\".format(splitX, splitY))

            Utility.makeDirectory(_cropDirectory)
            width, height = til.size  # Get dimensions
            _cropWidth = width / splitX
            _cropHeight = height / splitY
            print("width {} height {}".format(_cropWidth, _cropHeight))
            for x in range(splitX):
                for y in range(splitY):

                    _fullname = (
                        _cropDirectory
                        + _filename
                        + "_"
                        + colourKey
                        + "_x"
                        + str(x)
                        + "of"
                        + str(splitX)
                        + "y"
                        + str(y)
                        + "of"
                        + str(splitY)
                        + _extension
                    )
                    if skipIfFileExists and os.path.isfile(_fullname):
                        print(_fullname, "exists so not making it.")
                        return _fullname
                    left = x * _cropWidth
                    top = y * _cropHeight
                    right = left + _cropWidth
                    bottom = top + _cropHeight
                    _crop = til.crop((left, top, right, bottom))
                    _crop.save(_fullname)
                    print(
                        "finished crop. {} \nl{} t{} r{} b{}".format(
                            _fullname, left, top, right, bottom
                        )
                    )

        _fullname = os.path.join(_directory, _filename + _subname + "_" + colourKey + _extension)
        til.save(_fullname)
        print("finished making. " + _fullname)

    def getRingCircles(self, invertBG=False) -> [FCircle]:
        _proportion = FCircle.radiusOuterProportionRCircle
        if invertBG:
            _invertBG = not self.invertBG
        else:
            _invertBG = self.invertBG
        _mode = 0
        _rCircles = []
        _semitones = self.change.getSemitones()
        for st in range(12):
            _degree = st * 360 / 12 - 90
            _radian = _degree * 3.141596238 / 180

            if st in _semitones:
                if self.change.notes[0] == JazzNote("1"):
                    _colourTranspose = (st + self.colourTranspose - _semitones[0]) % 12
                else:
                    _colourTranspose = (
                        _semitones[0] + self.colourTranspose - _semitones[_mode]
                    ) % 12
                _rCircle = FCircle(
                    [
                        self.x
                        + math.cos(_radian) * self.r * _proportion / FCircle.proportion,
                        self.y
                        + math.sin(_radian) * self.r * _proportion / FCircle.proportion,
                    ],
                    self.r * FCircle.radiusOfEachRCircle / FCircle.proportion,
                    self.change.mode(_mode)
                    if self.change.notes[0] == JazzNote("1")
                    else self.change.getRotation(st - _semitones[0]),
                    _colourTranspose,
                    _proportion,
                    invertBG=_invertBG,
                )
                # print('stuuuuuuff',st +self.colourTranspose -_semitones[_mode])
                _rCircles.append(_rCircle)
                _mode += 1

        """for i in range(len(self.change):
            _rCircles.append(FCircle
                ([self.x,
                self.y],
                self.r,
                change.mode(i),
                self.colourTranspose+\
                self.change.notes[i].semitonesFromOne(),
                self.r*RCircle.proportion/FCircle.proportion)
                             )"""
        return _rCircles

    def getInnerCircles(
        self,
        defaultSize=0.618033988749894848204586834,
        useRingCircles=True,
        alternateBG=True,
        invertBG=False,
        useChordCircles=False,
        usePolyCircles=True,
        showDebug=False,
        circleType="FCircle",
    ) -> list[FCircle]:
        # .7071067811865475 1 over the square root of 2
        # .618033988749894848204586834 phi

        if invertBG:
            _invertBG = not self.invertBG
        else:
            _invertBG = self.invertBG

        if len(self.change.notes) == 0:
            return []

        # defaultSize contrrols the single note shape

        if False and len(self.change.notes) == 1:
            _newSize = self.r * defaultSize
            _angle = self.change.notes[0].getAngle(
                returnDegrees=False, goQuarterTurnAnticlockwise=True
            )

            """return [FCircle(centre=[self.x + math.cos(_angle) * (self.r - _newSize),
                                    self.y + math.sin(_angle) * (self.r - _newSize)],
                            r=_newSize, change=self.change,
                            colourTranspose=self.colourTranspose,
                            invertBG=_invertBG)]"""

            _artCircles = [
                FCircle(
                    centre=[
                        self.x + math.cos(_angle) * _newSize,
                        self.y + math.sin(_angle) * _newSize,
                    ],
                    r=_newSize,
                    change=Change([]),
                    colourTranspose=self.colourTranspose,
                    invertBG=_invertBG,
                )
            ]
        else:
            _artCircles = np.array([])
        _innerCircles = np.array([])
        _chordCrops = np.array([])
        _poly = np.array([])
        _ringCircles = np.array([])
        _circles = _artCircles
        if useChordCircles:
            # print('chord c')
            _chordCrops = self.getChordCrops(invertBG=True)
            # _circles += _chordCrops
            _circles = np.append(_circles, _chordCrops)
        if useRingCircles:
            # print('chord r')
            _ringCircles = self.getRingCircles(invertBG=False)
            _circles = np.append(_circles, _ringCircles)
            # _circles += _ringCircles
        if usePolyCircles and not len(self.change) in (1, 2):
            # print('chord p  {}'.format(_poly))
            _poly = self.getPolyCircle(invertBG=True)
            # input('poly {} FCircle {}'.format(_poly,self))
            # print('done chord p')
            if not _poly is None:
                _circles = np.append(_circles, _poly)
            else:
                raise TypeError("poly failed {}".format(_poly))
        # print('circles:',_circles)

        valid_circles = _circles[np.logical_and(np.not_equal(_circles, None),
                                                np.greater_equal([_circle.r * 2 for _circle in _circles],
                                                                 FCircle.minimumCirclePixels))]
        _innerCircles = np.append(_innerCircles, valid_circles)
        
        
        #Non Vectorised
        '''for circle in _circles:
            if circle is None:
                continue
            elif circle.r * 2 < FCircle.minimumCirclePixels:
                continue
            else:
                # _innerCircles.append(circle)
                _innerCircles = np.append(_innerCircles, circle)'''

        if showDebug:
            input(
                "helpful error.. \ncrops {}\npoly {}\nrings {}\ncircles {}\n innercircles {}".format(
                    len(_chordCrops),
                    _poly,
                    len(_ringCircles),
                    len(_circles),
                    len(_innerCircles),
                )
            )

        if len(_innerCircles) > 0 and type(_innerCircles[0]) != FCircle:
            input("error I think. the list should be flattened")
        # print('returning innercircles')
        return _innerCircles

    def getChordCrops(self, invertBG=False) -> list[FCircle]:
        if len(self.change.notes) <= 1:
            return []
        _bigR = self.r  # *FCircle.proportion
        _chordCrops = Graphics.getChordCentres(
            change=self.change,
            R=_bigR,
            returnFCircle=True,
            colourTranspose=self.colourTranspose,
            topLeftOrigin=False,
        )
        _chordCrops = [
            i.movePosition(self.x - 2 * self.r, self.y - 2 * self.r)
            for i in _chordCrops
        ]

        if not invertBG:
            for i in _chordCrops:
                i.invertBG = self.invertBG
            return _chordCrops
        else:
            for i in _chordCrops:
                i.invertBG = not self.invertBG
            return _chordCrops

    def getPolyCircle(
        self, invertBG=False, returnPoly=False, scale=1.0, shrinkIntoOrigin=True
    ) -> FCircle:
        # print('start getPolyCircle()')

        _poly = self.change.getPolygon(
            self.x - self.r,
            self.y - self.r,
            self.x + self.r,
            self.y + self.r,
            scale=scale,
            shrinkIntoOrigin=shrinkIntoOrigin,
            returnFCircle=not returnPoly,
            colourTranspose=self.colourTranspose,
        )

        # print('got _poly')
        if returnPoly:
            return _poly
        if _poly is None:
            return None

        # _poly = _poly.movePosition(self.x - 2 * self.r,
        #                             self.y - 2 * self.r)
        # _poly.movePosition(se)
        if invertBG:
            _poly.invertBG = not self.invertBG
        else:
            _poly.invertBG = self.invertBG
        return _poly

    def movePosition(self, xDist, yDist) -> FCircle:
        return FCircle(
            [self.x + xDist, self.y + yDist], self.r, self.change, self.colourTranspose
        )
from Graphics import Graphics


from __future__ import annotations, print_function
import os, sys, threading, glob
sys.path.append('../')
class Utility:
    numToCipher = {
        "1": "ONE",
        "2": "TWO",
        "3": "THREE",
        "4": "FOUR",
        "5": "FIVE",
        "6": "SIX",
        "7": "SEVEN",
        "8": "EIGHT",
        "9": "NINE",
    }




    projectClasses = (
        'wayofchange',
        'JazzNote',
        'Change'

    )

    @classmethod
    def grepProjectFiles(cls, search_str:str,extensions=('tex','py'),ignore_case=False,replace_str=None):
        #Utility.print('searching through',os.getcwd(),glob.glob('**/*.*',recursive=True))
        result = ''
        excluded_files = ['grailOfScale' + str(i) + '.py' for i in range(20, 24)]
        files_containing_str = []
        for root_dir in (
                os.path.abspath(os.path.join(os.path.abspath(__file__),'..','..')),
                os.path.abspath(os.path.join(os.path.abspath(__file__),'..','..','tex')),
                os.path.abspath(os.path.join(os.path.abspath(__file__), '..', '..', 'tex','waysExplained')),
                os.path.abspath(os.path.join(os.path.abspath(__file__),'..')),

        ):
            for extension in extensions:
                if root_dir == '.':
                    #input(os.listdir(os.path.join('..')))
                    filenames = [os.path.abspath(os.path.join(root_dir,f)) for f in os.listdir(os.path.abspath('../')) if f.lower().endswith('.' + extension.lower())]
                else:
                    filenames = glob.glob(root_dir + '/*.' + extension,recursive=True)
                    filenames = [os.path.abspath(os.path.join(root_dir,f)) for f in filenames]
                #filter
                filenames = [f for f in filenames if f.split('.')[-1] in extensions]
                Utility.print('searching through', root_dir,extension,filenames,)


                for filename in filenames:
                    #extension = filename.split('.')[-1]
                    if any([excluded_file in filename for excluded_file in excluded_files]):
                        continue
                    if os.path.isfile(filename):


                        #Utility.print("{} is a file so we should search for {}.".format(filename,search_str))

                        with open(filename,'r',encoding='utf8') as file:
                            try:
                                for i, line in enumerate(file):
                                    if ignore_case:
                                        search_str = search_str.lower()
                                        line = line.lower()
                                    if search_str in line:
                                        line_str = f'File "{filename}", line {i + 1}: {line}'
                                        Utility.print(line_str)
                                        result += line_str
                                        files_containing_str.append(filename)


                            except UnicodeDecodeError:
                                print(f'unicode error in {filename}')
                    else:
                        Utility.print("{} is not file so skipping.".format(filename))
        Utility.print(f'looking for {search_str}')
        print(result)
        files_containing_str = tuple(set(files_containing_str))
        if replace_str is not None and len(files_containing_str):
            if input(f'Replace "{search_str}" with "{replace_str}" in the files:\n{", ".join(files_containing_str)}? Y/Yes to do it. WARNING There is no undo!') in ('Y','y','Yes','yes'):
                print('doing the replace operation')
                for filename in set(files_containing_str):
                    Utility.replaceStrInFile(filename=filename,string=search_str, replacement=replace_str)
            else:
                print('not replacing because you did not type yes.')

        return result

    origInput = input
    origPrint = print
    @classmethod
    def print(cls, *args, sep=' ',end='\n'):
        from inspect import currentframe, getframeinfo
        frameinfo = getframeinfo(currentframe().f_back)
        linestr = 'File "{}", line {}'.format(frameinfo.filename, frameinfo.lineno)
        while len(linestr) < 60: linestr += ' '
        cls.origPrint(linestr + '\n', *args)

    origInput = input

    @classmethod
    def input(cls,*args, sep=' '):
        # print('it is finally hpapening with Utility.input({})'.format(s))
        # frameinfo = getframeinfo(currentframe())
        fmt_str = ' File "{}", line {}, in {}'
        abs_filepath = os.path.abspath(__file__)
        import inspect
        #lineno = currentframe().f_back.f_lineno
        try:
            frame = inspect.currentframe().f_back.f_back
            prevFrame = inspect.currentframe().f_back

            lineno = frame.f_lineno
            prevLineno= prevFrame.f_lineno
        except AttributeError:
            try:
                frame = inspect.currentframe().f_back
                prevFrame = inspect.currentframe()
                lineno = frame.f_lineno
                prevLineno = prevFrame.f_lineno
            except AttributeError:
                frame = inspect.currentframe()
                lineno = frame.f_lineno
                prevLineno = None
                prevFrame = None

        stack = inspect.stack()


        # currentframe().f_back.f_locals


        #line_str = fmt_str.format(abs_filepath, lineno, frame)
        fmt_str = 'File "{}", line {},  \nin: {},  \nat: {} \nFile "{}", line {},  '
        line_str = '\n'.join([
                fmt_str.format(
                str(f.filename),
                prevLineno,
                '{}({})'.format(str(f.frame).split(' code ')[-1][:-1],[k +'='+str(v)+', ' for k,v in f.frame.f_locals.items()]),
                f.code_context[0].split(' code ')[-1].strip(),
                str(f.filename),
                f.lineno,
                #f.code_context[0]
        ) \
                for f in stack if f.code_context and f.code_context[0].strip().startswith('input(' )])
        return Utility.origInput(
            #'\n' + line_str + '\n' + sep.join([str(s) for s in args]) + ' :: input() called from line ' + str(lineno))
            '\n' + line_str + '\n' + sep.join([str(s) for s in args]))


    @classmethod
    def semitoneDistToFreqDiff(cls,f1, f2):
        return 12 * (math.log(f2 / f1) / math.log(2))

    @classmethod
    def secondSmallest(cls,iterable):
        return sorted(set(iterable))[1]

    @classmethod
    def secondBiggest(cls, iterable):
        return sorted(set(iterable))[-2]
    @classmethod
    def removeNestedParentheses(cls,s,brackets=('(',')')):
        ret = ''
        skip = 0
        for i in s:
            if i == brackets[0]:
                skip += 1
            elif i == brackets[1] and skip > 0:
                skip -= 1
            elif skip == 0:
                ret += i
        return ret
    
    @classmethod
    def allIndexes(cls,iterable,value)->[int]:
        if type(iterable) == str:
            if value in ('',' '): raise ValueError('did you really want the indexes of all spaces in a the string? "{}"'.format(iterable))
            indexes = [index for index in range(len(iterable)) if iterable.startswith(value, index)]
            '''if len(indexes) > 2:
                Utility.input('\nman oh man value: "{}"\n{}\n{}'.format(value,indexes,iterable))'''
            return indexes
        else:
            return [index for index, i in  enumerate(iterable) if i == value]

    @classmethod
    def compileApp():
        import PyInstaller.__main__

        PyInstaller.__main__.run([
            'my_script.py',
            '--onefile',
            '--windowed'
        ])

    @classmethod
    def threaded(cls,function,args:[]=False):
        # Call work function
        if args:
            t1 = threading.Thread(target=function, args=args)
        else:
            t1 = threading.Thread(target=function)
        t1.start()


    @classmethod
    def getClassVariables(cls, Class):
        return {key: value for key, value in Class.__dict__.items() if not key.startswith('__') and not callable(key)}

    @classmethod
    def rasterisePDF(cls,path,greyScale=False,invertColour=False):

        newPath = path.replace(
            '.pdf','Rasterised.pdf'
        )
        print('begin converting',path,'to a rasterised .pdf! This may take a long time.')
        _startTime = datetime.datetime.now()
        pages = pdf2image.convert_from_path(path,transparent=True,size=(8.5*300,None),dpi=300,fmt='png',)
        #imTarget = Image.open("/Users/apple/Desktop/bbd.jpg")

        im_list = pages
        pdf1_filename = newPath
        im1 = im_list[0]

        #im1.save(pdf1_filename, "PDF", resolution=8.5*1200, save_all=True, append_images=im_list[1:],
        #         subsampling=0,optimize=True)
        print('it took',datetime.datetime.now()-_startTime)
        pdf = FPDF()
        # imagelist is the list with all image filenames
        for i,image in enumerate(tqdm(pages)):
            pageDir = os.path.join(Graphics.getDirectoryOfColourScheme(greyScale,invertColour),'ChordsOfChangeBookPages')
            pagePath = os.path.join(pageDir,str(i)+'.png')
            Utility.makeDirectory(pageDir)
            image.save(pagePath)
            pdf.add_page()
            pdf.image(pagePath, 0, 0, 210, 297,)
        pdf.output(newPath, "F")
        print('finished rasterizing '+pdf1_filename)

    @classmethod
    def makeNumberCipheredIfPreceededBy(cls, text: str, number, cipherTriggers):
        if cipherTriggers == None:
            cipherTriggers = (
                "0",
                "=",
                ".",
                Unicode.chars["Flat"],
                Unicode.chars["Sharp"],
            )


        _done = False
        _cipheredText = ""
        for c, char in enumerate(text):
            if char == str(number):
                if c != 0 and (
                    text[c - 1] in cipherTriggers
                    or (text[c - 2] in cipherTriggers and text[c - 1] == " ")
                ):
                    _cipheredText += cls.numToCipher[char]
                else:
                    _cipheredText += char
            else:
                _cipheredText += char
        return _cipheredText

    @classmethod
    def makeNumberUnciphered(cls, text, number: int = None):
        if number == None:
            numbers = tuple(range(10))
        else:
            numbers = (number,)
        for num in cls.numToCipher.keys():
            text = text.replace(cls.numToCipher[num], num)
        return text

    @classmethod
    def replaceTextInFile(cls, filename, searchStr, replaceStr):
        # Utility.input(type(file))
        # Read in the file
        if type(filename) != str:
            raise TypeError("hey file{}".format(str(file)))
        with open(filename, "r") as file:
            filedata = file.read()
        # Replace the target string
        filedata = filedata.replace(searchStr, replaceStr)

        # Write the file out again
        with open(filename, "w", encoding="UTF8") as file:
            file.write(filedata)

    @classmethod
    def getParameters(cls, func):
        keys = func.__code__.co_varnames[: func.__code__.co_argcount][::-1]
        sorter = {j: i for i, j in enumerate(keys[::-1])}
        values = func.__defaults__[::-1]
        kwargs = {i: j for i, j in zip(keys, values)}
        sorted_args = tuple(
            sorted([i for i in keys if i not in kwargs], key=sorter.get)
        )
        sorted_kwargs = {i: kwargs[i] for i in sorted(kwargs.keys(), key=sorter.get)}
        return sorted_args, sorted_kwargs

    @classmethod
    def saveToFile(cls, text: str, filePath):
        with open(filePath, "w+", encoding="utf-8") as file:
            file.write(text)
            file.close()
            print("wrote to ////file:{}".format(filePath))

    @classmethod
    def splitList(cls, l: [], n: int) -> [[]]:
        return np.array_split(l,n)
        '''matrix = []
        for i in range(n):
            start = round(len(l) / n * i)
            end = round(len(l) / n * (i + 1))
            matrix.append(l[start:end])
        return matrix'''

    @classmethod
    def removeAccentsFromText(cls, s: str):
        # Needs unicodedata
        return "".join(
            c
            for c in unicodedata.normalize("NFD", s)
            if unicodedata.category(c) != "Mn"
        )

    @classmethod
    def multiple_replace(cls, string, rep_dict):
        pattern = re.compile(
            "|".join([re.escape(k) for k in sorted(rep_dict, key=len, reverse=True)]),
            flags=re.DOTALL,
        )
        return pattern.sub(lambda x: rep_dict[x.group(0)], string)

    @classmethod
    def printPrettyVars(cls, *vars, sep=" ", indicator="="):
        """
        Gets the names and values of vars. Does it from the out most frame inner-wards.
        :param vars: variable to get name from.
        :param sep: what to put between vars.
        :param indicator: what to put between name and value.
        :return: string
        """
        for var in vars:
            for fi in reversed(inspect.stack()):
                names = [
                    var_name
                    for var_name, var_val in fi.frame.f_locals.items()
                    if var_val is var
                ]
                values = [
                    var_val
                    for var_name, var_val in fi.frame.f_locals.items()
                    if var_val is var
                ]
                if len(names) > 0 and names[0] != "var":
                    print(names[0], indicator, values[0].__repr__(), end=sep, sep="")

    @classmethod
    def openFile(cls, filename):
        try:
            if sys.platform == "win32":
                os.startfile(filename)
            else:
                opener = "open" if sys.platform == "darwin" else "xdg-open"
                subprocess.call([opener, filename])
            print("opening _thisDoc:", filename)
        except:
            raise OSError("we were not able to open {}.".format(filename))

    @classmethod
    def makeCharEscaped(cls, char, string):
        return string.replace(char)

    @classmethod
    def writeToFile(cls, data: str, filename):
        file = open(filename, "w+")
        file.write(data)
        file.close()

    @classmethod
    def makeStrCamelCase(cls, string, firstWordIsLower=True):
        string = "".join(x for x in string.title() if not x.isspace())
        if firstWordIsLower:
            string = string[0].lower() + string[1:]
        return string

    @classmethod
    def replaceStrInFile(cls, filename, string, replacement):
        # Read in the file
        with open(filename, "r", encoding="utf-8") as file:
            filedata = file.read()

        # Replace the target string
        filedata = filedata.replace(string, replacement)

        # Write the file out again
        with open(filename, "w", encoding="utf-8") as file:
            file.write(filedata)

    @classmethod
    def floatToFraction(cls, num, latexFractions=True):
        import math
        integerPart = math.floor(num)
        # At this part we will add making the 2/7 fraction etc
        numOrig = num
        num = round(num - integerPart, 6)
        if num == 0:
            _fractionUnicode = ""
            _latexUnicode = ""
        elif num == 0.5:
            _fractionUnicode = "½"
            _latexUnicode = "$\\frac{1}{2}$"
        elif num == 0.333333:
            _fractionUnicode = "⅓"
            _latexUnicode = "$\\frac{1}{3}$"
        elif num == 0.666667:
            _fractionUnicode = "⅔"
            _latexUnicode = "$\\frac{2}{3}$"
        elif num == 0.25:
            _fractionUnicode = "¼"
            _latexUnicode = "$\\frac{1}{4}$"
        elif num == 0.75:
            _fractionUnicode = "¾"
            _latexUnicode = "$\\frac{3}{4}$"
        elif num == 0.2:
            _fractionUnicode = "⅕"
            _latexUnicode = "$\\frac{1}{5}$"
        elif num == 0.4:
            _fractionUnicode = "⅖"
            _latexUnicode = "$\\frac{2}{5}$"
        elif num == 0.6:
            _fractionUnicode = "⅗"
            _latexUnicode = "$\\frac{3}{5}$"
        elif num == 0.8:
            _fractionUnicode = "⅘"
            _latexUnicode = "$\\frac{4}{5}$"
        elif num == 0.142857:
            _fractionUnicode = "⅐"
            _latexUnicode = "$\\frac{1}{7}$"
        elif num == 0.285714:
            _fractionUnicode = "²⁄₇"
            _latexUnicode = "$\\frac{2}{7}$"
        elif num == 0.428571:
            _fractionUnicode = "³⁄₇"
            _latexUnicode = "$\\frac{3}{7}$"
        elif num == 0.571429:
            _fractionUnicode = "⁴⁄₇"
            _latexUnicode = "$\\frac{4}{7}$"
        elif num == 0.714286:
            _fractionUnicode = "⁵⁄₇"
            _latexUnicode = "$\\frac{5}{7}$"
        elif num == 0.857143:
            _fractionUnicode = "⁶⁄₇"
            _latexUnicode = "$\\frac{6}{7}$"
        elif num == 0.111111:
            _fractionUnicode = "⅑"
            _latexUnicode = "$\\frac{1}{9}$"
        elif num == 0.222222:
            _fractionUnicode = "²⁄₉"
            _latexUnicode = "$\\frac{2}{9}$"
        elif num == 0.444444:
            _fractionUnicode = "⁴⁄₉"
            _latexUnicode = "$\\frac{4}{9}$"
        elif num == 0.555556:
            _fractionUnicode = "⁵⁄₉"
            _latexUnicode = "$\\frac{5}{9}$"
        elif num == 0.777778:
            _fractionUnicode = "⁷⁄₉"
            _latexUnicode = "$\\frac{7}{9}$"
        elif num == 0.888889:
            _fractionUnicode = "⁸⁄₉"
            _latexUnicode = "$\\frac{8}{9}$"
        elif num == 0.909091:
            _fractionUnicode = "10⁄11"
            _latexUnicode = "$\\frac{10}{11}$"
        else:
            # raise ValueError(numOrig)
            _fractionUnicode = str(round(num, 2))
            _latexUnicode = _fractionUnicode
        _str = ""
        if integerPart:
            _str += str(integerPart)
        if latexFractions:
            _str += _latexUnicode
        else:
            _str += _fractionUnicode
        return _str

    @classmethod
    def rgbToHex(cls, rgbs):
        # hsls = [(0.61, 0.43981, 0.2806), (0.69, 0.528195, 0.2277)]
        return ["".join("%02X" % round(i * 100) for i in rgb) for rgb in rgbs]

    @classmethod
    def draw_ellipse(cls, image, bounds, width=1, outline="white", antialias=4):
        """Improved ellipse drawing function, based on PIL.ImageDraw."""
        # https://stackoverflow.com/questions/32504246/draw-ellipse-in-python-pil-with-line-thickness
        # Use a single channel image (mode='L') as mask.
        # The size of the mask can be increased relative to the imput image
        # to get smoother looking results.
        mask = Image.new(
            size=[int(dim * antialias) for dim in image.size], mode="L", color="black"
        )
        draw = ImageDraw.Draw(mask)

        # draw outer shape in white (color) and inner shape in black (transparent)
        for offset, fill in (width / -2.0, "white"), (width / 2.0, "black"):
            left, top = [(value + offset) * antialias for value in bounds[:2]]
            right, bottom = [(value - offset) * antialias for value in bounds[2:]]
            draw.ellipse([left, top, right, bottom], fill=fill)

        # downsample the mask using PIL.Image.LANCZOS
        # (a high-quality downsampling filter).
        mask = mask.resize(image.size, Image.LANCZOS)
        # paste outline color to input image through the mask
        image.paste(outline, mask=mask)

    @classmethod
    def replaceStrInFilenames(
        cls, dir="C:\\Users\\Edrihan\\PycharmProjects\\wayofchange\\Graphics"
    ):
        for filepath in glob.iglob("./**/*.txt", recursive=True):
            with open(filepath, encoding="Latin-1") as file:
                s = file.read()
            s = s.replace("_", " ")
            with open(filepath, "w") as file:
                file.write(s)
                print("replacing char in {}".format(filepath))

    @classmethod
    def listIsSequential(cls, theList: [int], insertZero=True):
        first = theList[0]
        if insertZero and -1 in theList:
            # If the list contains [-1,1] in that order
            if theList.index(1) - theList.index(-1) == 1:
                theList = theList.insert(theList.index(1), 0)
        if all([c == (change - first) for c, change in enumerate(theList)]):
            return True
        else:
            return False

    @classmethod
    def turnLstToMatrix(cls, lst: list, cols: int = None, rows: int = None) -> [[]]:
        if rows == None and cols == None:
            raise ValueError("need to specify rows or columns")
        elif rows != None and cols != None:
            if rows * cols < len(lst):
                raise ValueError(
                    "not enough rows {} and cols {} to fit results.\n{}".format(
                        rows, cols, lst
                    )
                )
        elif rows != None:
            cols = math.floor(len(lst) / rows) + (len(lst) % rows != 0)
        elif cols != None:
            rows = math.floor(len(lst) / cols) + (len(lst) % cols != 0)
        _matrix = []
        _i = 0
        for row in range(rows):
            _matrix.append([])
            for col in range(cols):
                if _i < len(lst):
                    _matrix[-1].append(lst[_i])
                else:
                    _matrix[-1].append(None)

                _i += 1
        return _matrix



    @classmethod
    def flattenList(cls, lst: [list]) -> []:
        _flatten = lambda l: [item for sublist in l for item in sublist]
        if not isinstance(lst, (list,)):
            raise TypeError("{} is not a list.".format(lst))
        elif len(lst) >= 1 and not isinstance(lst[0], (list,)):
            return lst
        else:
            return _flatten(lst)

    @classmethod
    def flipAxesOfList(cls, lst: [list]):
        if not type(lst[0]) == list:
            raise TypeError("flipAxesOfList() accepts a list of lists.")
        _flippedLst = []
        for i in lst[0]:
            _flippedLst.append([1 for i in lst])
        for r, row in enumerate(lst):
            for d, data in enumerate(row):
                if (len(_flippedLst) + 1) < d:
                    _flippedList.append([])
                if (len(_flippedLst[d]) + 1) < r:
                    _flippedLst[d].append("")
                _flippedLst[d][r] = data
        # print(lst,_flippedLst,'blakfladkf')
        # Utility.input()
        return _flippedLst

    @classmethod
    def getFontStrings(cls):
        font_file = "/home/user/raleway.ttf"
        command = "otfinfo"
        params = ["--info"]
        result = subprocess.run(
            [command, *params, font_file], stdout=subprocess.PIPE
        ).stdout
        font_name_re = re.compile(r"Full name:\s*(.*)")
        font_name = font_name_re.findall(result.decode())
        print(font_name[0])

    @classmethod
    def makeDirectory(cls, path, printSuccess=True, printSkip=False):
        if os.path.isdir(path):
            if printSkip:
                print("skipped making {} because it already exists.".format(path))
            return True

        if 'F:' in path:
            raise TypeError('asdfasdf')
        os.makedirs(path, 0o777)
        if printSuccess:
            print("Successfully created the directory %s " % path)

    @classmethod
    def insertSubstringInStr(cls, pos: int, string: str, substring: str) -> str:
        return string[:pos] + substring + string[pos:]

    @classmethod
    def numPartOfStr(cls, string: str, includeDecimalPoint=True) -> float:
        _numPart = ""
        for char in string:
            if char.isdigit() or (includeDecimalPoint == True and char == "."):
                _numPart += char
        return float(_numPart)

    @classmethod
    def checkForDuplicates(cls, listToCheck: list):

        if len(set(listToCheck)) != len(listToCheck):
            Utility.input("woww, there are duplicates..")
        else:
            Utility.input("nice!- no duplicates!")
        print(
            "dupes",
            "\n  ".join([k for k, v in Counter(quals).items() if v > 1]),
            "length",
            len([k for k, v in Counter(quals).items() if v > 1]),
        )

        # indexes of dupes
        D = defaultdict(list)
        for i, item in enumerate(listToCheck):
            D[item].append(i)
        D = {k: v for k, v in D.items() if len(v) > 1}
        # Generic one
        print("D:", "\nd:".join([str(i[0]) + str(i[1]) for i in D.items()]))
        # specific one only for something I can't remember
        """print('D:', '\nd:'.join(
            [str(i[0]) + str([n + 1 for n in i[1]]) + str([str(theBook.sequence[a].straightenDegrees()) for a in i[1]])
             for i in D.items()]))"""

    @classmethod
    def allPinyin(cls):
        print("Hexagram pinyin\\\\")
        for i in Hexagram.pinyin:
            print(i)
        print("Trigram pinyin\\\\")
        print(Trigram.allPinyin())
        print("Tetragram pinyin\\\\")
        print(Tetragram.allPinyin())

input = Utility.input
print = Utility.print
def main():
    Utility.grepProjectFiles(search_str='FCircleChart', replace_str=None,extensions=('tex','py'))
if __name__ == "__main__":
    main()


class RCircle:
    circleBGColour = FCircle.circleBGColour
    canvasBGColour = FCircle.canvasBGColour
    lineColour = FCircle.lineColour
    minimumCirclePixels = FCircle.minimumCirclePixels
    # radiusOuterProportion = FCircle.proportion * 1.17
    radiusOuterProportion = FCircle.proportion * 1.3
    radiusOfEachCircle = 0.15


from Latex import Latex
from Utility import Utility
input = Utility.input
class Tetragram:
    # Tetragrams come from 太玄經 Tài Xuán Jīng
    # Characters from http://www.chinaknowledge.de/Literature/Daoists/taixuanjing.html
    # 首 Tetragrams Head/divisions

    # 樂 short for 樂經／乐经 (Yuèjīng, “Classic of Music”).
    allowedWays = [
        "symbol",
        "name",
        "chinese",
        "pinyin",
        "number",
        "subChange",
        "subpage",
    ]
    semitones = {}
    ternaryList = {}
    #book = Book(
    #    notesToUse=["b2", "2", "b3"],
    #    notesToInclude=["1"],
    #)

    # Digrams
    # ⚌    ⚎    𝌃    ⚍    ⚏    𝌁    𝌂    𝌄    𝌅
    # 1     2     6     4    5     7    3     8    9
    # None are the third bit
    semitones[(0, 1, 2, 3,)] = {
        "number": "1",
        "symbol": "𝌆",
        "name": "Centre",
        "pinyin": "zhōng",
        "chinese": "中",
        "subpage": 8,
    }
    semitones[(1, 2, 3,)] = {
        "number": "2",
        "symbol": "𝌇",
        "name": "Full-Circle",
        "pinyin": "zhōu",
        "chinese": "周",
        "subpage": -8,
    }
    ternaryList[(None, True, True, True)] = {
        "number": "3",
        "symbol": "𝌈",
        "name": "Mired",
        "pinyin": "xián",
        "chinese": "礥",
    }
    semitones[(0, 2, 3,)] = {
        "number": "4",
        "symbol": "𝌉",
        "name": "Barrier",
        "pinyin": "xián",
        "chinese": "閑",
        "simplified": "闲",
        "subpage": 7,
    }  # 閑
    semitones[(2, 3,)] = {
        "number": "5",
        "symbol": "𝌊",
        "name": "Keeping-Small",
        "pinyin": "shǎo",
        "chinese": "少",
        "subpage": -7,
    }
    ternaryList[(None, False, True, True)] = {
        "number": "6",
        "symbol": "𝌋",
        "name": "Contrariety",
        "pinyin": "lì",
        "chinese": "戾",
    }
    ternaryList[(True, None, True, True)] = {
        "number": "7",
        "symbol": "𝌌",
        "name": "Ascent",
        "pinyin": "shàng",
        "chinese": "上",
    }
    ternaryList[(True, None, True, True)] = {
        "number": "8",
        "symbol": "𝌍",
        "name": "Opposition",
        "pinyin": "gàn",
        "chinese": "幹",
    }
    ternaryList[(None, None, True, True)] = {
        "number": "9",
        "symbol": "𝌎",
        "name": "Branching",
        "pinyin": "shū",
        "chinese": "𤕠",
    }
    semitones[(0, 1, 3)] = {
        "number": "10",
        "symbol": "𝌏",
        "name": "Distortion",
        "pinyin": "xiàn",
        "chinese": "羨",
        "subpage": 6,
    }  # Defectiveness
    semitones[(1, 3)] = {
        "number": "11",
        "symbol": "𝌐",
        "name": "Divergence",
        "pinyin": "chà",
        "chinese": "差",
        "subpage": -6,
    }  # M.c. says like 'lack'
    ternaryList[(None, True, False, True)] = {
        "number": "12",
        "symbol": "𝌑",
        "name": "Youthfulness",
        "pinyin": "tóng",
        "chinese": "童",
    }
    semitones[(0, 3)] = {
        "number": "13",
        "symbol": "𝌒",
        "name": "Increase",
        "pinyin": "zēng",
        "chinese": "增",
        "subpage": 4,
    }
    semitones[(3,)] = {
        "number": "14",
        "symbol": "𝌓",
        "name": "Penetration",
        "pinyin": "ruì",
        "chinese": "銳",
        "subpage": -4,
    }
    ternaryList[(None, False, False, True)] = {
        "number": "15",
        "symbol": "𝌔",
        "name": "Reach",
        "pinyin": "dá",
        "chinese": "達",
        "simplified": "达",
    }
    ternaryList[(True, None, False, True)] = {
        "number": "16",
        "symbol": "𝌕",
        "name": "Contact",
        "pinyin": "jiāo",
        "chinese": "交",
    }
    ternaryList[(False, None, False, True)] = {
        "number": "17",
        "symbol": "𝌖",
        "name": "Holding Back",
        "pinyin": "ruǎn",
        "chinese": "耎",
    }
    ternaryList[(None, None, False, True)] = {
        "number": "18",
        "symbol": "𝌗",
        "name": "Waiting",
        "pinyin": "xī",
        "chinese": "徯",
    }
    ternaryList[(True, True, None, True)] = {
        "number": "19",
        "symbol": "𝌘",
        "name": "Following",
        "pinyin": "cóng",
        "chinese": "從",
    }
    ternaryList[(False, True, None, True)] = {
        "number": "20",
        "symbol": "𝌙",
        "name": "Advance",
        "pinyin": "jìn",
        "chinese": "進",
    }
    ternaryList[(None, True, None, True)] = {
        "number": "21",
        "symbol": "𝌚",
        "name": "Release",
        "pinyin": "shì",
        "chinese": "釋",
    }
    ternaryList[(True, False, None, True)] = {
        "number": "22",
        "symbol": "𝌛",
        "name": "Resistance",
        "pinyin": "gé",
        "chinese": "格",
    }
    ternaryList[(False, False, None, True)] = {
        "number": "23",
        "symbol": "𝌜",
        "name": "Ease",
        "pinyin": "yí",
        "chinese": "夷",
    }
    ternaryList[(None, False, None, True)] = {
        "number": "24",
        "symbol": "𝌝",
        "name": "Joy",
        "pinyin": "lè",
        "chinese": "樂",
    }  # Edrihan
    ternaryList[(True, None, None, True)] = {
        "number": "25",
        "symbol": "𝌞",
        "name": "Contention",
        "pinyin": "zhēng",
        "chinese": "爭",
    }
    ternaryList[(False, None, None, True)] = {
        "number": "26",
        "symbol": "𝌟",
        "name": "Endeavor",
        "pinyin": "wù",
        "chinese": "務",
    }
    ternaryList[(None, None, None, True)] = {
        "number": "27",
        "symbol": "𝌠",
        "name": "Duties",
        "pinyin": "shì",
        "chinese": "事",
    }
    semitones[(0, 1, 2)] = {
        "number": "28",
        "symbol": "𝌡",
        "name": "Change",
        "pinyin": "gēng",
        "chinese": "更",
        "subpage": 5,
    }  # jīng
    semitones[(1, 2)] = {
        "number": "29",
        "symbol": "𝌢",
        "name": "Decisiveness",
        "pinyin": "duàn",
        "chinese": "斷",
        "subpage": -5,
    }
    ternaryList[(None, True, True, False)] = {
        "number": "30",
        "symbol": "𝌣",
        "name": "Bold Resolution",
        "pinyin": "yì",
        "chinese": "毅",
    }
    semitones[(0, 2)] = {
        "number": "31",
        "symbol": "𝌤",
        "name": "Packing",
        "pinyin": "zhuāng",
        "chinese": "裝",
        "subpage": 3,
    }
    semitones[(2,)] = {
        "number": "32",
        "symbol": "𝌥",
        "name": "Legion",
        "pinyin": "zhòng",
        "chinese": "衆",
        "subpage": -3,
    }
    ternaryList[(None, False, True, False)] = {
        "number": "33",
        "symbol": "𝌦",
        "name": "Closeness",
        "pinyin": "mì",
        "chinese": "密",
    }
    ternaryList[(True, None, True, False)] = {
        "number": "34",
        "symbol": "𝌧",
        "name": "Kinship",
        "pinyin": "qīn",
        "chinese": "親",
    }
    ternaryList[(True, None, True, False)] = {
        "number": "35",
        "symbol": "𝌨",
        "name": "Gathering",
        "pinyin": "liǎn",
        "chinese": "斂",
    }  # liàn
    ternaryList[(None, None, True, False)] = {
        "number": "36",
        "symbol": "𝌩",
        "name": "Strength",
        "pinyin": "qiáng",
        "chinese": "強",
    }  # qiǎng jiàng
    semitones[(0, 1,)] = {
        "number": "37",
        "symbol": "𝌪",
        "name": "Purity",
        "pinyin": "suì",
        "chinese": "睟",
        "subpage": 2,
    }
    semitones[(1,)] = {
        "number": "38",
        "symbol": "𝌫",
        "name": "Fullness",
        "pinyin": "chéng",
        "chinese": "盛",
        "subpage": -2,
    }  # shèng
    ternaryList[(None, None, True, False)] = {
        "number": "39",
        "symbol": "𝌬",
        "name": "Residence",
        "pinyin": "jū",
        "chinese": "居",
    }
    semitones[(0,)] = {
        "number": "40",
        "symbol": "𝌭",
        "name": "Law",
        "pinyin": "fǎ",
        "chinese": "法",
        "subpage": 1,
    }  # Model  fǎ, fā, fá, fà
    semitones[()] = {
        "number": "41",
        "symbol": "𝌮",
        "name": "Response",
        "pinyin": "yīng",
        "chinese": "應",
        "subpage": -1,
    }  # yìng
    ternaryList[(None, False, False, False)] = {
        "number": "42",
        "symbol": "𝌯",
        "name": "Going to Meet",
        "pinyin": "yù",
        "chinese": "遇",
    }
    ternaryList[(True, None, False, False)] = {
        "number": "43",
        "symbol": "𝌰",
        "name": "Encounters",
        "pinyin": "yíng",
        "chinese": "迎",
    }
    # Note that in traditional Chinese (Taiwan), this character is classified as a variant form of 灶. (44)
    ternaryList[(False, None, False, False)] = {
        "number": "44",
        "symbol": "𝌱",
        "name": "Stove",
        "pinyin": "zào",
        "chinese": "竈",
        "simplified": "灶",
    }
    """ dài
    Used in 大夫 (dàifu, “doctor”).
    Used in 大城(Dàichéng, “Daicheng, Hebei”).
    Used in 大王(dàiwáng, “(in operas, old novels)
    king; ringleader”)."""
    ternaryList[(None, None, False, False)] = {
        "number": "45",
        "symbol": "𝌲",
        "name": "Greatness",
        "pinyin": "dà",
        "chinese": "大",
    }
    ternaryList[(True, True, None, False)] = {
        "number": "46",
        "symbol": "𝌳",
        "name": "Enlargement",
        "pinyin": "kuò",
        "chinese": "廓",
    }
    ternaryList[(False, True, None, False)] = {
        "number": "47",
        "symbol": "𝌴",
        "name": "Pattern",
        "pinyin": "wén",
        "chinese": "文",
    }
    ternaryList[(None, True, None, False)] = {
        "number": "48",
        "symbol": "𝌵",
        "name": "Ritual",
        "pinyin": "lǐ",
        "chinese": "禮",
    }
    ternaryList[(True, False, None, False)] = {
        "number": "49",
        "symbol": "𝌶",
        "name": "Flight",
        "pinyin": "táo",
        "chinese": "逃",
    }
    ternaryList[(False, False, None, False)] = {
        "number": "50",
        "symbol": "𝌷",
        "name": "Vastness",
        "pinyin": "táng",
        "chinese": "唐",
    }  # Wasting
    ternaryList[(None, False, None, False,)] = {
        "number": "51",
        "symbol": "𝌸",
        "name": "Constancy",
        "pinyin": "cháng",
        "chinese": "常",
    }
    ternaryList[(True, None, None, False)] = {
        "number": "52",
        "symbol": "𝌹",
        "name": "Measure",
        "pinyin": "dù",
        "chinese": "度",
    }
    ternaryList[(False, None, None, False)] = {
        "number": "53",
        "symbol": "𝌺",
        "name": "Eternity",
        "pinyin": "yǒng",
        "chinese": "永",
    }
    ternaryList[(None, None, None, False)] = {
        "number": "54",
        "symbol": "𝌻",
        "name": "Unity",
        "pinyin": "kūn",
        "chinese": "昆",
    }
    ternaryList[(True, True, True, None)] = {
        "number": "55",
        "symbol": "𝌼",
        "name": "Diminishment",
        "pinyin": "jiǎn",
        "chinese": "減",
    }
    ternaryList[(False, True, True, None)] = {
        "number": "56",
        "symbol": "𝌽",
        "name": "Closed Mouth",
        "pinyin": "jìn",
        "chinese": "唫",
    }
    ternaryList[(None, False, False, None)] = {
        "number": "57",
        "symbol": "𝌾",
        "name": "Guardedness",
        "pinyin": "shǒu",
        "chinese": "守",
    }
    ternaryList[(True, False, True, None)] = {
        "number": "58",
        "symbol": "𝌿",
        "name": "Gathering In",
        "pinyin": "xī",
        "chinese": "翕",
    }
    ternaryList[(False, False, True, None)] = {
        "number": "59",
        "symbol": "𝍀",
        "name": "Massing",
        "pinyin": "jù",
        "chinese": "聚",
    }
    ternaryList[(None, False, True, None)] = {
        "number": "60",
        "symbol": "𝍁",
        "name": "Accumulation",
        "pinyin": "jī",
        "chinese": "積",
    }
    ternaryList[(True, None, True, None)] = {
        "number": "61",
        "symbol": "𝍂",
        "name": "Embellishment",
        "pinyin": "shì",
        "chinese": "飾",
    }
    ternaryList[(False, None, True, None)] = {
        "number": "62",
        "symbol": "𝍃",
        "name": "Doubt",
        "pinyin": "yí",
        "chinese": "疑",
    }  # níng
    ternaryList[(None, None, True, None)] = {
        "number": "63",
        "symbol": "𝍄",
        "name": "Watch",
        "pinyin": "shì",
        "chinese": "視",
    }
    ternaryList[(True, True, False, None)] = {
        "number": "64",
        "symbol": "𝍅",
        "name": "Sinking",
        "pinyin": "shěn",
        "chinese": "沈",
    }  # Cenkner
    ternaryList[(False, True, False, None)] = {
        "number": "65",
        "symbol": "𝍆",
        "name": "Inner",
        "pinyin": "nèi",
        "chinese": "內",
    }
    ternaryList[(None, True, False, None)] = {
        "number": "66",
        "symbol": "𝍇",
        "name": "Departure",
        "pinyin": "qù",
        "chinese": "去",
    }
    ternaryList[(True, False, False, None)] = {
        "number": "67",
        "symbol": "𝍈",
        "name": "Darkening",
        "pinyin": "huì",
        "chinese": "晦",
    }
    ternaryList[(False, False, False, None)] = {
        "number": "68",
        "symbol": "𝍉",
        "name": "Dimming",
        "pinyin": "méng",
        "chinese": "瞢",
    }
    ternaryList[(None, False, False, None)] = {
        "number": "69",
        "symbol": "𝍊",
        "name": "Exhaustion",
        "pinyin": "qióng",
        "chinese": "窮",
    }
    ternaryList[(True, None, False, None)] = {
        "number": "70",
        "symbol": "𝍋",
        "name": "Severence",
        "pinyin": "gē",
        "chinese": "割",
    }
    ternaryList[(False, None, False, None)] = {
        "number": "71",
        "symbol": "𝍌",
        "name": "Stoppage",
        "pinyin": "zhǐ",
        "chinese": "止",
    }
    ternaryList[(None, None, False, None)] = {
        "number": "72",
        "symbol": "𝍍",
        "name": "Hardness",
        "pinyin": "jiān",
        "chinese": "堅",
    }
    ternaryList[(True, True, None, None,)] = {
        "number": "73",
        "symbol": "𝍎",
        "name": "Completion",
        "pinyin": "chéng",
        "chinese": "成",
    }
    ternaryList[(False, True, None, None,)] = {
        "number": "74",
        "symbol": "𝍏",
        "name": "Closure",
        "pinyin": "zhì",
        "chinese": "䦯",
    }
    ternaryList[(None, True, None, None,)] = {
        "number": "75",
        "symbol": "𝍐",
        "name": "Failure",
        "pinyin": "shī",
        "chinese": "失",
    }
    ternaryList[(True, False, None, None,)] = {
        "number": "76",
        "symbol": "𝍑",
        "name": "Aggravation",
        "pinyin": "jù",
        "chinese": "劇",
    }  # jí
    ternaryList[(False, False, None, None,)] = {
        "number": "77",
        "symbol": "𝍒",
        "name": "Compliance",
        "pinyin": "",
        "chinese": "馴",
    }
    ternaryList[(None, False, None, None,)] = {
        "number": "78",
        "symbol": "𝍓",
        "name": "Verging",
        "pinyin": "xùn",
        "chinese": "將",
    }  # xún On the verge
    ternaryList[(True, None, None, None,)] = {
        "number": "79",
        "symbol": "𝍔",
        "name": "Difficulties",
        "pinyin": "nán",
        "chinese": "難",
    }  # nàn
    ternaryList[(False, None, None, None,)] = {
        "number": "80",
        "symbol": "𝍕",
        "name": "Labouring",
        "pinyin": "qín",
        "chinese": "勤",
    }
    ternaryList[(False, None, None, None,)] = {
        "number": "81",
        "symbol": "𝍖",
        "name": "Fostering",
        "pinyin": "yǎng",
        "chinese": "養",
    }

    if Latex.on and Latex.replaceBuggyPinyinChars:
        for i in semitones:
            for replacement in Latex.pinyinReplacements:
                semitones[i]["pinyin"] = semitones[i]["pinyin"].replace(
                    replacement, Latex.pinyinReplacements[replacement]
                )
        for i in ternaryList:
            for replacement in Latex.pinyinReplacements:
                ternaryList[i]["pinyin"] = ternaryList[i]["pinyin"].replace(
                    replacement, Latex.pinyinReplacements[replacement]
                )

    @classmethod
    def makeTable(
        cls, arrangePositionsByWay=True, flipAxes=False, rewriteFile=True
    ) -> str:
        printWays = Tetragram.allowedWays
        _str = ""
        rows = []
        for t, tetragram in enumerate(Tetragram.semitones):
            rows.append([])
            for way in printWays:
                if way in Tetragram.semitones[tetragram].keys():
                    # rows[-1].append(Tetragram.notesets[tetragram][way])
                    rows[-1].append(
                        Change.makeFromSet(tetragram).getTetragram(
                            [way], decorateWithSmallCircle=True
                        )[0]
                    )
                else:
                    # print(Change.isValidWay('Change Number'))
                    # input('else' + way)

                    if way == "subChange":
                        way = "Change Number"
                    if Change.isValidWay(way):
                        # input('is valid change way')
                        for t, transposition in enumerate((0, 4, 8)):
                            _change = Change.makeFromSet(
                                [s + transposition for s in tetragram]
                            )
                            rows[-1].append(_change.byWays(way))
                    else:
                        raise TypeError("{} is not a validWay of Change".format(way))

        if flipAxes:
            rows = Utility.flipAxesOfList(rows)
        _str += Latex.makeTabular(rows)
        if rewriteFile:
            _filename = Project.tetragramTableFilename
            text_file = open(_filename, "w")
            text_file.write("%s" % _str)
            text_file.close()
            print("rewrote", _filename)
        print(_str)
        input("thar be rows")

    @classmethod
    def allPinyin(cls):
        for i in cls.semitones.keys():
            print(cls.semitones[i]["pinyin"])
        for i in cls.ternaryList.keys():
            print(cls.ternaryList[i]["pinyin"])